import os
import six
from six.moves import map
from slapos.util import rmtree

def createFolder(folder, clean=False):
  if os.path.exists(folder):
    if not clean:
      return
    rmtree(folder)
  os.mkdir(folder)

if six.PY3:
  def deunicodeData(data):
    return data
else:
  def deunicodeData(data):
    if isinstance(data, list):
      return list(map(deunicodeData, data))
    if isinstance(data, unicode):
      return data.encode('utf8')
    if isinstance(data, dict):
      return {deunicodeData(key): deunicodeData(value)
              for key, value in six.iteritems(data)}
    return data



from __future__ import print_function

import datetime
import json
import traceback
import time
import requests
from functools import wraps
from uritemplate import expand

import slapos.slap
from slapos.slap import SoftwareProductCollection
from slapos.client import SOFTWARE_PRODUCT_NAMESPACE
from slapos.util import xml2dict
from requests.exceptions import HTTPError
from ..taskdistribution import SAFE_RPC_EXCEPTION_LIST
from . import logger

import six

# max time to instance changing state: 3 hour
MAX_INSTANCE_TIME = 60*60*3

SOFTWARE_STATE_UNKNOWN = "SOFTWARE_STATE_UNKNOWN"
SOFTWARE_STATE_INSTALLING = "SOFTWARE_STATE_INSTALLING"
SOFTWARE_STATE_INSTALLED = "SOFTWARE_STATE_INSTALLED"
SOFTWARE_STATE_DESTROYING = "SOFTWARE_STATE_DESTROYING"

INSTANCE_STATE_UNKNOWN = "INSTANCE_STATE_UNKNOWN"
INSTANCE_STATE_STARTING = "INSTANCE_STATE_STARTING"
INSTANCE_STATE_STARTED = "started"
INSTANCE_STATE_STARTED_WITH_ERROR = "INSTANCE_STATE_STARTED_WITH_ERROR"
INSTANCE_STATE_STOPPING = "INSTANCE_STATE_STOPPING"
INSTANCE_STATE_STOPPED = "stopped"
INSTANCE_STATE_DESTROYED = "destroyed"

TESTER_STATE_INITIAL = "TESTER_STATE_INITIAL"
TESTER_STATE_NOTHING = "TESTER_STATE_NOTHING"
TESTER_STATE_SOFTWARE_INSTALLED = "TESTER_STATE_SOFTWARE_INSTALLED"
TESTER_STATE_INSTANCE_INSTALLED = "TESTER_STATE_INSTANCE_INSTALLED"
TESTER_STATE_INSTANCE_STARTED = "TESTER_STATE_INSTANCE_STARTED"
TESTER_STATE_INSTANCE_UNINSTALLED = "TESTER_STATE_INSTANCE_UNINSTALLED"


# Simple decorator to prevent raise due small
# network failures.
def retryOnNetworkFailure(func,
    _except_list = SAFE_RPC_EXCEPTION_LIST + (
      HTTPError, slapos.slap.ConnectionError),
    ):
  def wrapper(*args, **kw):
    retry_time = 64
    while True:
      try:
        return func(*args, **kw)
      except _except_list:
        traceback.print_exc()

      print('Network failure. Retry method %s in %i seconds' % (func, retry_time))
      time.sleep(retry_time)
      retry_time = min(retry_time*1.5, 640)

  return wraps(func)(wrapper)


class SlapOSMasterCommunicator(object):
  latest_state = None

  def __init__(self, slap, slap_supply, slap_order, url):
    self.slap = slap
    self.slap_order = slap_order
    self.slap_supply = slap_supply
    self.hateoas_navigator = self.slap._hateoas_navigator

    if url is not None and \
      url.startswith(SOFTWARE_PRODUCT_NAMESPACE):

      product = SoftwareProductCollection(logger, self.slap)
      try:
        url = product.__getattr__(url[len(SOFTWARE_PRODUCT_NAMESPACE):])
      except AttributeError as e:
        logger.warning('Error on get software release: %s ', e.message)

    self.url = url  

  #########################################################
  # Wrapper functions to support network retries
  #########################################################

  @retryOnNetworkFailure
  def _supply(self, state="available"):
    if self.computer_guid is None:
      logger.info('Nothing to supply for %s.', self.name)
      return None
    logger.info('Supply %s@%s', self.url, self.computer_guid)
    return self.slap_supply.supply(self.url, self.computer_guid, state)

  @retryOnNetworkFailure
  def _request(self, state, instance_title=None, request_kw=None, shared=False, software_type="RootSoftwareInstance"):
    if instance_title is not None:
      self.name = instance_title 
    if request_kw is not None:
      if isinstance(request_kw, bytes):
        self.request_kw = json.loads(request_kw.decode('utf-8'))
      elif isinstance(request_kw, six.text_type):
        self.request_kw = json.loads(request_kw)
      else:
        self.request_kw = request_kw
    if self.request_kw is None:
      self.request_kw = {}
    logger.info('Request %s@%s: %s', self.url, self.name, state)
    self.latest_state = state
    return self.slap_order.request(
            software_release=self.url,
            partition_reference=self.name,
            shared=shared,
            state=state,
            **self.request_kw)

  @retryOnNetworkFailure
  def isInstanceRequested(self, instance_title):
    return len(self.hateoas_navigator._getInstanceTreeList(
        title=instance_title))

  @retryOnNetworkFailure
  def getComputer(self, reference):
    return self.hateoas_navigator.getComputer(reference)

  @retryOnNetworkFailure
  def getSoftwareInstallationList(self, computer_guid=None):
    return self.hateoas_navigator.getSoftwareInstallationList(computer_guid=computer_guid)

  @retryOnNetworkFailure
  def getInstanceUrlList(self):
    return self.hateoas_navigator.getInstanceTreeInstanceList(
      self.name)

  @retryOnNetworkFailure
  def getInformationFromInstance(self, url):
    return self.hateoas_navigator.jio_get(url)

  @retryOnNetworkFailure
  def _getSoftwareState(self, computer_guid=None):
    if self.computer_guid is None:
      return SOFTWARE_STATE_INSTALLED

    message = self.hateoas_navigator.getSoftwareInstallationNews(
            computer_guid, self.url)
    logger.info(message)
    if message.startswith("#error no data found"):
      return SOFTWARE_STATE_UNKNOWN

    if message.startswith('#access software release'):
      return SOFTWARE_STATE_INSTALLED

    if message.startswith('#error'):
      return SOFTWARE_STATE_INSTALLING

    return SOFTWARE_STATE_UNKNOWN

  @retryOnNetworkFailure
  def getRSSEntryFromMonitoring(self, base_url):
    if base_url is None:
      return {}

    feed_url = base_url + '/monitor-public/rssfeed.html'
    d = feedparser.parse(feed_url)

    if len(d.entries) > 0:
      return {"date": d.entries[0].published,
              "message": d.entries[0].description,
              "title" : d.entries[0].title}
    return {}

  @retryOnNetworkFailure
  def _getInstanceState(self):
    latest_state = self.latest_state
    logger.info('latest_state = %r', latest_state)

    if latest_state is None:
      return INSTANCE_STATE_UNKNOWN

    message_list = []
    try:
      for instance in self.getInstanceUrlList():
        # we need to explicitly encode as utf-8 the unicode string we get
        instance["text_content"] = instance["text_content"].encode('utf8')
        news = instance['SoftwareInstance_getNewsDict']
        state = INSTANCE_STATE_UNKNOWN
        monitor_information_dict = {}

        is_slave = instance['portal_type'] == "Slave Instance"
        if is_slave:
          if len(instance['getConnectionXmlAsDict']) > 0:
            state =  INSTANCE_STATE_STARTED
        else:
          # not slave
          instance_state = news
          if instance_state.get('created_at', '-1') != "-1":
            instance_text = instance_state['text']

            if instance_text.startswith('#access Instance correctly started'):
              state =  INSTANCE_STATE_STARTED
            elif instance_text.startswith('#access Instance correctly stopped'):
              state =  INSTANCE_STATE_STOPPED
            elif instance_text.startswith('#destroy'):
              state = INSTANCE_STATE_DESTROYED
            elif instance_text.startswith('#error'):
              state = INSTANCE_STATE_STARTED_WITH_ERROR

        if state == INSTANCE_STATE_STARTED_WITH_ERROR:
          # search for monitor url
          monitor_v6_url = instance['getConnectionXmlAsDict'].get("monitor_v6_url")
          try:
            monitor_information_dict = self.getRSSEntryFromMonitoring(monitor_v6_url)
          except Exception:
            logger.exception('Unable to download promises for: %s', instance["title"])
            logger.error(traceback.format_exc())
            monitor_information_dict = {"message": "Unable to download"}

        instance["connection_dict"] = instance["getConnectionXmlAsDict"]
        instance["parameter_dict"] = xml2dict(instance["text_content"])
        message_list.append({
          'title': instance["title"],
          'slave': is_slave,
          'news': news,
          'information': instance,
          'monitor': monitor_information_dict,
          'state': state
        })

    except slapos.slap.ServerError:
      logger.error('Got an error requesting partition for its state')
      return INSTANCE_STATE_UNKNOWN
    except Exception:
      logger.exception("ERROR getting instance state")
      return INSTANCE_STATE_UNKNOWN

    started = 0
    stopped = 0
    self.message_history.append(message_list)
    for instance in message_list:
      if not instance['slave']:
        if instance['state'] in (INSTANCE_STATE_UNKNOWN, INSTANCE_STATE_STARTED_WITH_ERROR):
          return instance['state']
        elif instance['state'] == INSTANCE_STATE_STARTED:
          started = 1
        elif instance['state'] == INSTANCE_STATE_STOPPED:
          stopped = 1
      if instance['slave'] and instance['state'] == INSTANCE_STATE_UNKNOWN:
        return instance['state']

    if started and stopped:
      return INSTANCE_STATE_STOPPED

    if started:
      return INSTANCE_STATE_STARTED

    if stopped:
      return INSTANCE_STATE_STOPPED

  @retryOnNetworkFailure
  def _waitInstance(self, instance_title, state, max_time=MAX_INSTANCE_TIME):
    """
    Wait for 'max_time' an instance specific state
    """
    logger.info("Waiting for instance state: %s", state)
    start_time = time.time()
    while (not self._getInstanceState() == state
           and (max_time > (time.time()-start_time))):
      logger.info("Instance(s) not in %s state yet.", state)
      logger.info("Current state: %s", self._getInstanceState())
      time.sleep(15)
    if (time.time()-start_time) > max_time:
      error_message = "Instance '%s' not '%s' after %s seconds" %(instance_title, state, str(time.time()-start_time))
      return {'error_message' : error_message}
    logger.info("Instance correctly '%s' after %s seconds.", state, time.time() - start_time)
    return {'error_message' : None}

class SlapOSTester(SlapOSMasterCommunicator):

  def __init__(self,
              name,
              slap,
              slap_order,
              slap_supply,
              url, # software release url
              computer_guid=None, # computer for supply if desired
              request_kw=None
          ):
    super(SlapOSTester, self).__init__(
      slap, slap_supply, slap_order, url)

    self.name = name
    self.computer_guid = computer_guid

    if isinstance(request_kw, bytes):
      self.request_kw = json.loads(request_kw.decode('utf-8'))
    elif isinstance(request_kw, six.text_type):
      self.request_kw = json.loads(request_kw)
    else:
      self.request_kw = request_kw
    self.message_history = []

  def getInfo(self):
    info = ""
    info += "Software Release URL: %s\n" % (self.url)
    if self.computer_guid is not None:
      info += "Supply requested on: %s\n" % (self.computer_guid)
    info += "Instance Requested (Parameters): %s\n" % self.request_kw
    return info

  def supply(self, software_path=None, computer_guid=None, state="available"):
    if software_path is not None:
      self.url = software_path
    if computer_guid is not None:
      self.computer_guid = computer_guid
    self._supply(state)

  def requestInstanceStart(self, instance_title=None, request_kw=None, shared=False, software_type="RootSoftwareInstance"):
    self.instance = self._request(INSTANCE_STATE_STARTED, instance_title, request_kw, shared, software_type)

  def requestInstanceStop(self, instance_title=None, request_kw=None, shared=False, software_type="RootSoftwareInstance"):
    self.instance = self._request(INSTANCE_STATE_STOPPED, instance_title, request_kw, shared, software_type)

  def requestInstanceDestroy(self, instance_title=None, request_kw=None, shared=False):
    if not instance_title:
      instance_title = self.name
    self.destroyInstance(instance_title)

  def waitInstanceStarted(self, instance_title):
    error_message = self._waitInstance(instance_title, INSTANCE_STATE_STARTED)["error_message"]
    if error_message is not None:
      logger.error(error_message)
      logger.error("Instance '%s' will be stopped and test aborted.", instance_title)
      self.requestInstanceStop()
      time.sleep(60)
      raise ValueError(error_message)

  def waitInstanceStopped(self, instance_title):
    error_message = self._waitInstance(instance_title, INSTANCE_STATE_STOPPED)["error_message"]
    if error_message is not None:
      logger.error(error_message)
      raise ValueError(error_message)

  def waitInstanceDestroyed(self, instance_title):
    error_message = self._waitInstance(instance_title, INSTANCE_STATE_DESTROYED)["error_message"]
    if error_message is not None:
      logger.error(error_message)
      raise ValueError(error_message)

  def getMasterFrontendDict(self):
    def getInstanceGuid():
      try:
        return self.instance.getInstanceGuid()
      except Exception:
        return None
    frontend_master_ipv6 = None
    instance_guid = None
    for instance in self.getInstanceUrlList():
      if instance["title"] == "Monitor Frontend apache-frontend-1":
        try:
          frontend_master_ipv6 = xml2dict(instance['text_content'])['url']
        except Exception as e:
          pass
    start_time = time.time()
    while not getInstanceGuid() and time.time()-start_time < 60*5:
      sleep(60)
    return {'instance_guid' : getInstanceGuid(), 'frontend_master_ipv6' : frontend_master_ipv6}

  # XXX TODO
  # In the future, this should allow customization so each project to be tested parses its own information,
  # probably in the test suite definition class
  def getInstanceUrlDict(self):
    user = None
    password = None
    frontend_url_list = []
    for instance in self.getInstanceUrlList():
      if "frontend-" in instance["title"]:
        try:
          frontend = [instance["title"].replace("frontend-", ""),
                      instance["getConnectionXmlAsDict"]["secure_access"]]
          frontend_url_list.append(frontend)
        except Exception as e:
          logger.info("Frontend url not generated yet for instance: " + instance["title"])
          pass
      if instance["title"] == self.name:
        try:
          connection_json = json.loads(instance["getConnectionXmlAsDict"]["_"])
          user = connection_json["inituser-login"]
          password = connection_json["inituser-password"]
        except Exception as e:
          raise ValueError("user and password not found in connection parameters. Error while instantiating?")
    return {'user' : user, 'password' : password, 'frontend-url-list' : frontend_url_list }

  def destroyInstance(self, instance_title):
    self.name = instance_title
    instance_url_list = self.getInstanceUrlList()
    if instance_url_list:
      for instance in instance_url_list:
        if instance["title"] != instance_title:
          self._request(INSTANCE_STATE_DESTROYED, instance["title"])
        else:
          root_instance = instance
      logger.info("Going to destroy root partition: " + str(instance_title))
      self._request(INSTANCE_STATE_DESTROYED, root_instance["title"])
    else:
      logger.info("Instance not found")

class SoftwareReleaseTester(SlapOSTester):
  deadline = None

  def __init__(self,
              name,
              slap,
              slap_order,
              slap_supply,
              url, # software release url
              computer_guid=None, # computer for supply if desired
              request_kw=None,
              software_timeout=3600,
              instance_timeout=3600,
          ):
    super(SoftwareReleaseTester, self).__init__(
      name, slap, slap_order, slap_supply, url, computer_guid, request_kw)

    self.state = TESTER_STATE_INITIAL
    self.transition_dict = {
      # step function
      # delay
      # next_state
      # software_state
      # instance_state
      TESTER_STATE_INITIAL: (
        lambda t: None,
        None,
        TESTER_STATE_NOTHING,
        None,
        None,
      ),
      TESTER_STATE_NOTHING: (
        lambda t: t._supply("available"),
        int(software_timeout),
        request_kw is None and TESTER_STATE_INSTANCE_UNINSTALLED or \
            TESTER_STATE_SOFTWARE_INSTALLED,
        SOFTWARE_STATE_INSTALLED,
        None,
      ),
      TESTER_STATE_SOFTWARE_INSTALLED: (
        lambda t: t._request(INSTANCE_STATE_STARTED),
        int(instance_timeout),
        TESTER_STATE_INSTANCE_STARTED,
        None,
        INSTANCE_STATE_STARTED,
      ),
      TESTER_STATE_INSTANCE_STARTED: (
        lambda t: t._request(INSTANCE_STATE_DESTROYED),
        int(1200),
        TESTER_STATE_INSTANCE_UNINSTALLED,
        None,
        INSTANCE_STATE_STOPPED,
      ),
      TESTER_STATE_INSTANCE_UNINSTALLED: (
        lambda t: t._supply(INSTANCE_STATE_DESTROYED),
        int(1200),
        None,
        None,
        None,
      ),
     }

  def __repr__(self):
      deadline = self.deadline
      if deadline is not None:
          deadline -= time.time()
          deadline = '+%is' % (deadline, )
      return '<%s(state=%s, deadline=%s) at %x>' % (
          self.__class__.__name__, self.state, deadline, id(self))

  def getFormatedLastMessage(self):
    if len(self.message_history) == 0:
      return "No message"

    summary = "Summary about the test. Instance List and Status:\n"
    message = "Last information about the tester:\n"
    if self.message_history[-1] is not None:
      message_list = self.message_history[-1]
      for entry in message_list:
        summary += "%s %s -> %s\n" % (
          entry['title'], entry["slave"] and "(slave)" or "", entry['state'])
        for prop in entry:
          if prop != "information":
            message += "%s = %s\n" % (prop, json.dumps(entry[prop], indent=2))
          
        message += "=== connection_dict === \n%s\n" % (
          json.dumps(entry["information"]["connection_dict"], indent=2))
        message += "\n"
        message += "=== parameter_dict === \n%s\n" % (
          json.dumps(entry["information"]["parameter_dict"], indent=2))
        message += "\n"
      message += "="*79
      message += "\n\n\n"
 
    return summary + message

  @retryOnNetworkFailure
  def teardown(self):
    """
    Interrupt a running test sequence, putting it in idle state.
    """
    logger.info('Invoking TearDown for %s@%s', self.url, self.name)
    if self.request_kw is not None:
       self._request(INSTANCE_STATE_DESTROYED)
    if self.computer_guid is not None:
      self._supply(INSTANCE_STATE_DESTROYED)
    self.state = TESTER_STATE_INSTANCE_UNINSTALLED

  def tic(self, now):
    """
    Check for missed deadlines (-> test failure), conditions for moving to
    next state, and actually moving to next state (executing its payload).
    """
    logger.debug('TIC')
    deadline = self.deadline

    if deadline < now and deadline is not None:
      raise TestTimeout(self.state)

    _, _, next_state, software_state, instance_state = self.transition_dict[
        self.state]

    if (software_state is None or
          software_state == self._getSoftwareState()) and (
          instance_state is None or
          instance_state == self._getInstanceState()):

      logger.debug('Going to state %s (%r)', next_state, instance_state)
      if next_state is None:
        return None

      self.state = next_state
      stepfunc, delay, _, _, _ = self.transition_dict[next_state]
      self.deadline = now + delay
      stepfunc(self)
    return self.deadline



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import os
import glob
import json
import logging

from . import logger
from .ProcessManager import SubprocessError, format_command
from .SlapOSControler import SlapOSControler
from .Utils import createFolder
from slapos.grid.utils import md5digest

def dealShebang(run_test_suite_path):
  with open(run_test_suite_path) as f:
    if f.read(2) == '#!':
      return f.readline().split(None, 1)
  return []

class UnitTestRunner(object):

  def __init__(self, testnode):
    self.testnode = testnode

  def _getSlapOSControler(self, working_directory, use_local_shared_part):
    """
    Create a SlapOSControler
    """
    return SlapOSControler(
               working_directory,
               self.testnode.config,
               use_local_shared_part=use_local_shared_part)

  def _prepareSlapOS(self, working_directory, slapos_instance,
          create_partition=1, software_path_list=None, use_local_shared_part=False, **kw):
    """
    Launch slapos to build software and partitions
    """
    slapproxy_log = os.path.join(self.testnode.config['log_directory'],
                                  'slapproxy.log')
    logger.debug('Configured slapproxy log to %r', slapproxy_log)
    reset_software = slapos_instance.retry_software_count > 10
    if reset_software:
      slapos_instance.retry_software_count = 0
    logger.info('testnode, retry_software_count: %r',
             slapos_instance.retry_software_count)

    # XXX Create a new controler because working_directory can be
    #     Different depending of the preparation
    slapos_controler = self._getSlapOSControler(
        working_directory,
        use_local_shared_part)

    slapos_controler.initializeSlapOSControler(slapproxy_log=slapproxy_log,
       process_manager=self.testnode.process_manager, reset_software=reset_software,
       software_path_list=software_path_list)
    self.testnode.process_manager.supervisord_pid_file = os.path.join(\
         slapos_controler.instance_root, 'var', 'run', 'supervisord.pid')
    method_list= ["runSoftwareRelease"]
    if create_partition:
      method_list.append("runComputerPartition")
    for method_name in method_list:
      slapos_method = getattr(slapos_controler, method_name)
      logger.debug("Before status_dict = slapos_method(...)")
      status_dict = slapos_method(self.testnode.config,
                                  environment=self.testnode.config['environment'],
                                  **kw)
      logger.info(status_dict)
      logger.debug("After status_dict = slapos_method(...)")
      if status_dict['status_code'] != 0:
         slapos_instance.retry = True
         slapos_instance.retry_software_count += 1
         raise SubprocessError(status_dict)
      else:
         slapos_instance.retry_software_count = 0
    return status_dict

  def prepareSlapOSForTestNode(self, test_node_slapos):
    """
    We will build slapos software needed by the testnode itself,
    like the building of selenium-runner by default
    """
    # report-url, report-project and suite-url are required to seleniumrunner
    # instance. This is a hack which must be removed.
    config = self.testnode.config
    return self._prepareSlapOS(test_node_slapos.working_directory,
              test_node_slapos, create_partition=0,
              software_path_list=config.get("software_list"),
              cluster_configuration={
                'report-url': config.get("report-url", ""),
                'report-project': config.get("report-project", ""),
                'suite-url': config.get("suite-url", ""),
              })

  def prepareSlapOSForTestSuite(self, node_test_suite):
    """
    Build softwares needed by testsuites.
    """
    return self._prepareSlapOS(node_test_suite.working_directory,
              node_test_suite,
              software_path_list=[node_test_suite.custom_profile_path],
              cluster_configuration={'_': json.dumps(node_test_suite.cluster_configuration)},
              use_local_shared_part=True)

  def getInstanceRoot(self, node_test_suite):
    return self._getSlapOSControler(
      node_test_suite.working_directory, True).instance_root

  def runTestSuite(self, node_test_suite, portal_url):
    config = self.testnode.config
    run_test_suite_path_list = glob.glob(
        self.getInstanceRoot(node_test_suite) + "/*/bin/runTestSuite")
    try:
      run_test_suite_path = min(run_test_suite_path_list)
    except ValueError:
      raise ValueError('No runTestSuite provided in installed partitions.')
    # Deal with Shebang size limitation
    invocation_list = dealShebang(run_test_suite_path)
    invocation_list += (run_test_suite_path,
      '--master_url', portal_url,
      '--revision', node_test_suite.revision,
      '--test_node_title', config['test_node_title'],
      '--test_suite', node_test_suite.test_suite,
      '--test_suite_title', node_test_suite.test_suite_title)
    soft = config['slapos_directory'] + '/soft/'
    software_list = [soft + md5digest(x) for x in config['software_list']]
    PATH = os.getenv('PATH', '')
    PATH = ':'.join(x + '/bin' for x in software_list) + (PATH and ':' + PATH)
    SLAPOS_TEST_SHARED_PART_LIST = os.pathsep.join(
        self._getSlapOSControler(
            node_test_suite.working_directory,
            True
        ).shared_part_list)
    SLAPOS_TEST_LOG_DIRECTORY = node_test_suite.log_folder_path
    supported_parameter_set = set(self.testnode.process_manager
      .getSupportedParameterList(run_test_suite_path))
    def path(name, compat): # BBB
        path, = filter(os.path.exists, (base + relative
          for relative in ('/bin/' + name, '/parts/' + compat)
          for base in software_list))
        return path
    for option, value in (
        ('--firefox_bin', lambda: path('firefox', 'firefox/firefox-slapos')),
        ('--frontend_url', lambda: config['frontend_url']),
        ('--node_quantity', lambda: config['node_quantity']),
        ('--xvfb_bin', lambda: path('xvfb', 'xserver/bin/Xvfb')),
        ('--project_title', lambda: node_test_suite.project_title),
        ('--shared_part_list', lambda: SLAPOS_TEST_SHARED_PART_LIST),
        ('--log_directory', lambda: SLAPOS_TEST_LOG_DIRECTORY),
        ):
      if option in supported_parameter_set:
        invocation_list += option, value()

    # TODO : include testnode correction ( b111682f14890bf )
    if hasattr(node_test_suite,'additional_bt5_repository_id'):
      additional_bt5_path = os.path.join(
              node_test_suite.working_directory,
              node_test_suite.additional_bt5_repository_id)
      invocation_list.extend(["--bt5_path", additional_bt5_path])
    # From this point, test runner becomes responsible for updating test
    # result. We only do cleanup if the test runner itself is not able
    # to run.
    createFolder(node_test_suite.test_suite_directory, clean=True)

    # Log the actual command with root logger
    root_logger = logging.getLogger()
    root_logger.info(
        "Running test suite with: %s",
        format_command(*invocation_list, PATH=PATH))

    def hide_distributor_url(s):
      # type: (bytes) -> bytes
      return s.replace(portal_url.encode('utf-8'), b'$DISTRIBUTOR_URL')

    self.testnode.process_manager.spawn(*invocation_list, PATH=PATH,
                          SLAPOS_TEST_SHARED_PART_LIST=SLAPOS_TEST_SHARED_PART_LIST,
                          SLAPOS_TEST_LOG_DIRECTORY=SLAPOS_TEST_LOG_DIRECTORY,
                          cwd=node_test_suite.test_suite_directory,
                          log_prefix='runTestSuite',
                          output_replacers=(hide_distributor_url,),
                          get_output=False)

  def getRelativePathUsage(self):
    """
    Used by the method testnode.constructProfile() to know
    if the software.cfg have to use relative path or not.
    """
    return False


##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import os
import psutil
import re
import subprocess
import threading
import signal
import sys
import time
from . import logger

MAX_TIMEOUT = 3600 * 4

class SubprocessError(EnvironmentError):
  def __init__(self, status_dict):
    self.status_dict = status_dict
  def __getattr__(self, name):
    return self.status_dict[name]
  def __str__(self):
    return 'Error %i' % self.status_code

class TimeoutError(EnvironmentError):
  def __init__(self):
    pass
  def __str__(self):
    return 'Timeout expired. Process killed'

class CancellationError(EnvironmentError):
  pass

_format_command_search = re.compile("[[\\s $({?*\\`#~';<>&|]").search
_format_command_escape = lambda s: "'%s'" % r"'\''".join(s.split("'"))

def format_command(*args, **kw):
  cmdline = []
  for k, v in sorted(kw.items()):
    if _format_command_search(v):
      v = _format_command_escape(v)
    cmdline.append('%s=%s' % (k, v))
  for v in args:
    if _format_command_search(v):
      v = _format_command_escape(v)
    cmdline.append(v)
  return ' '.join(cmdline)

def subprocess_capture(p, log_prefix, get_output=True, output_replacers=()):
  log = logger.info
  if log_prefix:
    log_prefix += ': '
  def readerthread(input, buffer):
    while True:
      data = input.readline()
      if not data:
        break
      for replacer in output_replacers:
        data = replacer(data)
      if get_output:
        buffer.append(data)
      log(log_prefix + (data if str is bytes else
                        data.decode('utf-8', errors='replace')).rstrip('\n'))
  if p.stdout:
    stdout = []
    stdout_thread = threading.Thread(target=readerthread,
                                     args=(p.stdout, stdout))
    stdout_thread.daemon = True
    stdout_thread.start()
  if p.stderr:
    stderr = []
    stderr_thread = threading.Thread(target=readerthread,
                                     args=(p.stderr, stderr))
    stderr_thread.daemon = True
    stderr_thread.start()
  p.wait()
  if p.stdout:
    stdout_thread.join()
  if p.stderr:
    stderr_thread.join()
  return (p.stdout and b''.join(stdout),
          p.stderr and b''.join(stderr))

def killCommand(pid):
  """
  To prevent processes from reacting to the KILL of other processes,
  we STOP them all first, and we repeat until the list of children does not
  change anymore. Only then, we KILL them all.
  """
  try:
    process = psutil.Process(pid)
    process.suspend()
    process_list = [process]
    new_list = process.children(recursive=True)
  except psutil.Error as e:
    return
  while new_list:
    process_list += new_list
    for child in new_list:
      try:
        child.suspend()
      except psutil.Error as e:
        logger.debug("killCommand/suspend: %s", e)
    time.sleep(1)
    try:
      new_list = set(process.children(recursive=True)).difference(process_list)
    except psutil.Error as e:
      new_list = []
  for process in process_list:
    try:
      process.kill()
    except psutil.Error as e:
      logger.debug("killCommand/kill: %s", e)

class ProcessManager(object):

  def __init__(self, max_timeout=MAX_TIMEOUT):
    self.process_pid_set = set()
    signal.signal(signal.SIGTERM, self.sigterm_handler)
    self.under_cancellation = False
    self.p = None
    self.result = None
    self.max_timeout = max_timeout
    self.timer_set = set()

  def spawn(self, *args, **kw):
    def timeoutExpired(p):
      if p.poll() is None:
        logger.warning('PROCESS TOO LONG OR DEAD, GOING TO BE TERMINATED')
        killCommand(p.pid)
        raise SubprocessError('Dead or too long process killed')

    if self.under_cancellation:
      raise CancellationError("Test Result was cancelled")
    get_output = kw.pop('get_output', True)
    log_prefix = kw.pop('log_prefix', '')
    new_session = kw.pop('new_session', True)
    subprocess_kw = {}

    output_replacers = kw.pop('output_replacers', ())
    cwd = kw.pop('cwd', None)
    if cwd:
      subprocess_kw['cwd'] = cwd
    if new_session:
      subprocess_kw['preexec_fn'] = os.setsid
    raise_error_if_fail = kw.pop('raise_error_if_fail', True)
    env = dict(os.environ, **kw) if kw else None
    command = format_command(*args, **kw)
    # obfuscate secrets from command, assuming command is utf-8
    for replacer in output_replacers:
      command = replacer(command.encode('utf-8')).decode('utf-8')
    logger.info('subprocess_kw : %r', subprocess_kw)
    logger.info('$ %s', command)
    sys.stdout.flush()

    with open(os.devnull) as stdin:
      p = subprocess.Popen(args, stdin=stdin, stdout=subprocess.PIPE,
                          stderr=subprocess.PIPE, env=env, close_fds=True,
                          **subprocess_kw)
      self.process_pid_set.add(p.pid)
      timer = threading.Timer(self.max_timeout, timeoutExpired, args=(p,))
      self.timer_set.add(timer)
      timer.start()
      stdout, stderr = subprocess_capture(p, log_prefix, get_output=get_output, output_replacers=output_replacers)
      timer.cancel()
      self.timer_set.discard(timer)
      result = dict(status_code=p.returncode, command=command,
                    stdout=stdout, stderr=stderr)
      self.process_pid_set.discard(p.pid)
      p.stdout.close()
      p.stderr.close()

      if self.under_cancellation:
        raise CancellationError("Test Result was cancelled")
      if raise_error_if_fail and p.returncode:
        raise SubprocessError(result)
    return result

  def getSupportedParameterList(self, program_path):
    # type: (str) -> Sequence[str]
    return (parameter.decode('utf-8') for parameter in
        re.findall(br'^  (--\w+)', self.spawn(program_path, '--help')['stdout'], re.M))

  def killall(self, path):
    """
    Kill processes of given name, only if they're orphan or subprocesses of
    the testnode.
    """
    to_kill_list = []
    pid = os.getpid()
    for process in psutil.process_iter():
      try:
        if not(path in str(process.cmdline())):
          continue
      except (psutil.AccessDenied, psutil.NoSuchProcess):
        continue
      logger.debug('ProcesssManager, killall on %s having pid %s',
               process.name, process.pid)
      to_kill_list.append(process.pid)
    for pid in to_kill_list:
      killCommand(pid)

  def killPreviousRun(self, cancellation=False):
    logger.debug('ProcessManager killPreviousRun, going to kill %r',
             self.process_pid_set)
    if cancellation:
      self.under_cancellation = True
    for timer in self.timer_set:
      timer.cancel()
    for pgpid in tuple(self.process_pid_set):
      killCommand(pgpid)
    try:
      pid_file = self.supervisord_pid_file
    except AttributeError:
      pass
    else:
      del self.supervisord_pid_file
      try:
        if os.path.exists(pid_file):
          with open(pid_file) as f:
            pid = int(f.read().strip())
          logger.debug('ProcessManager killPreviousRun,'
                       ' going to kill supervisor with pid %r', pid)
          os.kill(pid, signal.SIGTERM)
          # Give at most two minutes to supervisord to stop everything
          i = 0
          while True:
            try:
              psutil.Process(pid)
              logger.debug('ProcessManager killPreviousRun, supervisor still there with pid %r', pid)
            except psutil.NoSuchProcess:
              break
            else:
              time.sleep(1)
              i += 1
              if i > 120:
                logger.debug('ProcessManager killPreviousRun, supervisor still there after two minutes %r', pid)
                break
      except Exception:
        raise
        logger.exception(
          'ProcessManager killPreviousRun, exception when killing supervisor')
    self.process_pid_set.clear()

  def sigterm_handler(self, signal, frame):
    logger.debug('SIGTERM_HANDLER')
    sys.exit(1)



##############################################################################
#
# Copyright (c) 2011 Nexedi SA and Contributors. All Rights Reserved.
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsibility of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# guarantees and support are strongly advised to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 3
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
import os
import json
import time
import logging
from six.moves.urllib.parse import urljoin
from contextlib import contextmanager
from slapos.slap.slap import ConnectionError
from slapos.util import rmtree
from . import logger, log_formatter
from .ProcessManager import SubprocessError, ProcessManager, CancellationError
from subprocess import CalledProcessError
from .Updater import Updater
from .NodeTestSuite import NodeTestSuite, SlapOSInstance
from .ScalabilityTestRunner import ScalabilityTestRunner
from .UnitTestRunner import UnitTestRunner
from .Utils import deunicodeData
from .. import taskdistribution

MAX_TEMP_TIME = 0.01 # time in days we should keep temp files

PROFILE_PATH_KEY = 'profile_path'

test_type_registry = {
  'UnitTest': UnitTestRunner,
  'ScalabilityTest': ScalabilityTestRunner,
}

class TestNode(object):

  def __init__(self, config, max_temp_time=MAX_TEMP_TIME):
    self.config = config or {}
    self.process_manager = ProcessManager()
    self.working_directory = config['working_directory']
    self.node_test_suite_dict = {}
    self.file_handler = None
    self.max_log_time = float(config.get('keep_log_days', 15))
    self.max_temp_time = max_temp_time
    self.url_access = "https://[0::0]:0123" # Ipv6 + port of the node

  def purgeOldTestSuite(self, test_suite_data):
    reference_set = set(os.listdir(self.working_directory))
    reference_set.difference_update(x['test_suite_reference']
                                    for x in test_suite_data)
    for reference in reference_set:
      fpath = os.path.join(self.working_directory, reference)
      self.node_test_suite_dict.pop(reference, None)
      logger.info("testnode.purgeOldTestSuite, DELETING : %r", fpath)
      if os.path.isdir(fpath):
        rmtree(fpath)
      else:
        os.remove(fpath)
  
  def getNodeTestSuite(self, reference):
    try:
      node_test_suite = self.node_test_suite_dict[reference]
    except KeyError:
      node_test_suite = self.node_test_suite_dict[reference] = \
        NodeTestSuite(reference, self.working_directory)
      config = self.config
      node_test_suite.edit(log_directory=config['log_directory'])
    return node_test_suite

  def constructProfile(self, node_test_suite, test_type, use_relative_path=False):
    assert len(node_test_suite.vcs_repository_list), "we must have at least one repository"
    software_config_path = None
    profile_content_list = []
    revision_dict = dict(node_test_suite.revision_list)
    self.config['repository_path_list'] = []
    for vcs_repository in node_test_suite.vcs_repository_list:
      buildout_section_id = vcs_repository.get('buildout_section_id')
      repository_path = vcs_repository['repository_path']
      self.config['repository_path_list'].append(repository_path)
      try:
        profile_path = vcs_repository[PROFILE_PATH_KEY]
      except KeyError:
        pass
      else:
        if software_config_path is not None:
          raise ValueError(PROFILE_PATH_KEY + ' defined more than once')

        # Absolute path to relative path
        software_config_path = os.path.join(repository_path, profile_path)
        if use_relative_path :
          from_path = os.path.join(self.working_directory,
                                    node_test_suite.reference)
          software_config_path = os.path.relpath(software_config_path, from_path)

      # Construct sections
      if not(buildout_section_id is None):
        # Absolute path to relative
        if use_relative_path:
          from_path = os.path.join(self.working_directory,
                                    node_test_suite.reference)
          repository_path = os.path.relpath(repository_path, from_path)

        if test_type=="ScalabilityTest":
          profile_content_list.append("""
[%(buildout_section_id)s]
repository = %(url)s
revision = %(revision)s
ignore-ssl-certificate = true
develop = false
shared = true
""" %     {'url': vcs_repository['url'],
           'buildout_section_id': buildout_section_id,
          'revision': revision_dict[buildout_section_id][1]})
        else:
          profile_content_list.append("""
[%(buildout_section_id)s]
repository = %(repository_path)s
branch = %(branch)s
revision =
develop = false
shared = true
""" %     {'buildout_section_id': buildout_section_id,
          'repository_path' : repository_path,
          'branch' : vcs_repository.get('branch','master')})
    if software_config_path is None:
      raise ValueError(PROFILE_PATH_KEY + ' not defined')
    profile_content_list.sort()
    # Write file
    with open(node_test_suite.custom_profile_path, 'w') as f:
      f.write("[buildout]\nextends = %s\n%s" % (
        software_config_path,
        ''.join(profile_content_list)))

  def updateRevisionList(self, node_test_suite):
    config = self.config
    revision_list = []
    try:
      for vcs_repository in node_test_suite.vcs_repository_list:
        repository_path = vcs_repository['repository_path']
        repository_id = vcs_repository['repository_id']
        branch = vcs_repository.get('branch')
        # Make sure we have local repository
        updater = Updater(repository_path, git_binary=config['git_binary'],
           branch=branch, process_manager=self.process_manager,
           working_directory=node_test_suite.working_directory,
           url=vcs_repository["url"])
        updater.checkout()
        revision_list.append((repository_id, updater.getRevision()))
    except SubprocessError as error:
      # only limit to particular error, if we run that code for all errors,
      # then if server having most repositories is down for some time, we would
      # erase all repositories and facing later hours of downloads
      if b'index' in getattr(error, 'stderr', b''):
        rmtree(repository_path)
      logger.warning("Error while getting repository, ignoring this test suite",
                     exc_info=1)
      return False
    node_test_suite.revision_list = revision_list
    return True

  @contextmanager
  def suiteLog(self, node_test_suite):
    suite_log_path, folder_id = node_test_suite.createSuiteLog()
    handler = logging.FileHandler(filename=suite_log_path)
    handler.setFormatter(log_formatter)
    logger.info('Suite logfile: %s', suite_log_path)
    try:
      logger.propagate = False
      logger.addHandler(handler)
      yield folder_id
    finally:
      logger.propagate = True
      logger.removeHandler(handler)
      handler.close()

  def checkRevision(self, test_result, node_test_suite):
    if node_test_suite.revision == test_result.revision:
      return lambda: None
    logger.info('Disagreement on tested revision, checking out: %r != %r',
        node_test_suite.revision, test_result.revision)
    updater_kw = dict(git_binary=self.config['git_binary'],
                      process_manager=self.process_manager)
    updater_list = []
    revision_list = []
    for i, revision in enumerate(test_result.revision.split(',')):
      vcs_repository = node_test_suite.vcs_repository_list[i]
      repository_path = vcs_repository['repository_path']
      count, revision = revision.split('=')[1].split('-')
      revision_list.append((vcs_repository['repository_id'],
                            (int(count), revision)))
      # other testnodes on other boxes are already ready to test another
      # revision
      updater = Updater(repository_path, revision=revision, **updater_kw)
      updater.checkout()
      updater.git_update_server_info()
      updater.git_create_repository_link()
      updater_list.append(updater)
    node_test_suite.revision_list = revision_list
    def git_gc_auto():
      for updater in updater_list:
        updater.git_gc_auto()
    return git_gc_auto

  def _cleanupLog(self):
    log_directory = self.config['log_directory']
    prune_time = time.time() - 86400 * self.max_log_time
    for log_folder in os.listdir(log_directory):
      folder_path = os.path.join(log_directory, log_folder)
      if os.path.isdir(folder_path):
        if os.stat(folder_path).st_mtime < prune_time:
          logger.debug("deleting log directory %r", folder_path)
          rmtree(folder_path)

  def _cleanupTemporaryFiles(self):
    """
    buildout seems letting files under /tmp. To avoid regular error of
    missing disk space, remove old logs
    """
    temp_directory = self.config["system_temp_folder"]
    user_id = os.geteuid()
    prune_time = time.time() - 86400 * self.max_temp_time
    for temp_folder in os.listdir(temp_directory):
      folder_path = os.path.join(temp_directory, temp_folder)
      if (temp_folder.startswith("tmp") or
          temp_folder.startswith("buildout")):
        try:
          stat = os.stat(folder_path)
          if stat.st_uid == user_id and stat.st_mtime < prune_time:
            logger.debug("deleting temp directory %r", folder_path)
            if os.path.isdir(folder_path):
              rmtree(folder_path)
            else:
              os.remove(folder_path)
        except OSError:
          logger.warning("_cleanupTemporaryFiles exception", exc_info=1)

  def cleanUp(self):
    logger.debug('Testnode.cleanUp')
    self.process_manager.killPreviousRun()
    self.process_manager.killall(self.working_directory)
    self.process_manager.killall(self.config['slapos_directory'])
    self._cleanupLog()
    self._cleanupTemporaryFiles()

  def run(self):
    config = self.config
    portal_url = config['test_suite_master_url']
    test_node_slapos = SlapOSInstance(config['slapos_directory'])
    testnode_software_successfully_built = False
    try:
      while True:
        test_result = None
        try:
          node_test_suite = None
          self.cleanUp()
          begin = time.time()
          taskdistributor = taskdistribution.TaskDistributor(
              portal_url, logger=logger)
          self.taskdistribution = taskdistributor
          node_configuration = taskdistributor.subscribeNode(
            node_title=config['test_node_title'],
            computer_guid=config['computer_id'])
          if type(node_configuration) is str:
            # Backward compatiblity
            node_configuration = json.loads(node_configuration)
          if node_configuration is not None and \
              'process_timeout' in node_configuration \
              and node_configuration['process_timeout'] is not None:
            process_timeout = node_configuration['process_timeout']
            logger.info('Received and using process timeout from master: %i',
              process_timeout)
            self.process_manager.max_timeout = process_timeout
          test_suite_data = taskdistributor.startTestSuite(
            node_title=config['test_node_title'],
            computer_guid=config['computer_id'])
          if type(test_suite_data) is str:
            # Backward compatiblity
            test_suite_data = json.loads(test_suite_data)
          test_suite_data = deunicodeData(test_suite_data)
          logger.info("Got following test suite data from master : %r",
              test_suite_data)
          try:
            my_test_type = taskdistributor.getTestType()
          except Exception:
            logger.warning("testnode, error during requesting getTestType()"
              " method from the distributor.")
            raise
          # Select runner according to the test type
          try:
            runner_class = test_type_registry[my_test_type]
          except KeyError:
            logger.warning("testnode, Runner type %s not implemented.",
                           my_test_type)
            raise NotImplementedError
          runner = runner_class(self)
          logger.info("Type of current test is %s", my_test_type)
          # master testnode gets test_suites, slaves get nothing
          if not(testnode_software_successfully_built):
            testnode_software_status_dict = runner.prepareSlapOSForTestNode(test_node_slapos)
            if testnode_software_status_dict['status_code'] == 0:
              testnode_software_successfully_built = True
              logger.info("Will now skip build of testnode software")
          # Clean-up test suites
          self.purgeOldTestSuite(test_suite_data)
          for test_suite in test_suite_data:
            node_test_suite = self.getNodeTestSuite(
               test_suite.pop("test_suite_reference"))
            node_test_suite.edit(**test_suite)

            # XXX: temporary hack to prevent empty test_suite
            if not hasattr(node_test_suite, 'test_suite'):
              node_test_suite.edit(test_suite='')
            # kill processes from previous loop if any
            self.process_manager.killPreviousRun()
            if not self.updateRevisionList(node_test_suite):
              continue
            test_result = taskdistributor.createTestResult(
                     node_test_suite.revision, [],
                     config['test_node_title'], False,
                     node_test_suite.test_suite_title,
                     node_test_suite.project_title)
            logger.info("testnode, test_result : %r", test_result)
            if test_result is None:
              self.cleanUp() # XXX not a good place to do that
              continue
            with self.suiteLog(node_test_suite) as suite_log_folder_name:
              test_result.reportStatus(
                  'LOG url', urljoin(config['log_frontend_url'], suite_log_folder_name), '')
              git_gc_auto = self.checkRevision(test_result, node_test_suite)
              node_test_suite.edit(test_result=test_result)
              # get cluster configuration for this test suite, this is needed to
              # know slapos parameters to user for creating instances
              logger.info("Getting configuration from test suite %s",
                  node_test_suite.test_suite_title)
              generated_config = taskdistributor.generateConfiguration(
                node_test_suite.test_suite_title)
              json_data = json.loads(generated_config)
              cluster_configuration = deunicodeData(json_data['configuration_list'][0])
              node_test_suite.edit(cluster_configuration=cluster_configuration)
              # Now prepare the installation of SlapOS and create instance
              self.constructProfile(node_test_suite, my_test_type,
                                    runner.getRelativePathUsage())
              status_dict = runner.prepareSlapOSForTestSuite(node_test_suite)
              # SR built successfully. Any cloned repository (with --shared)
              # should be at the same revision, so it is safe to prune orphan
              # objects now.
              git_gc_auto()
              # Give some time so computer partitions may start
              # as partitions can be of any kind we have and likely will never have
              # a reliable way to check if they are up or not ...
              time.sleep(20)
              # XXX: Do not switch according to the test type. IOW, the
              #      following code must be moved to the test type class.
              if my_test_type == 'UnitTest':
                runner.runTestSuite(node_test_suite, portal_url)
              elif my_test_type == 'ScalabilityTest':
                error_message = None
                # A problem is appeared during runTestSuite
                if status_dict['status_code'] == 1:
                  error_message = "Software installation too long or error(s) are present during SR install."
                else:
                  status_dict = runner.runTestSuite(node_test_suite, portal_url)
                  # A problem is appeared during runTestSuite
                  if status_dict['status_code'] == 1:
                    error_message = status_dict['error_message']

                # If an error is appeared
                if error_message:
                  test_result.reportFailure(
                      stdout=error_message
                  )
                  logger.error(error_message)
                  raise ValueError(error_message)
              else:
                raise NotImplementedError
              # break the loop to get latest priorities from master
              break
        except (SubprocessError, CalledProcessError, ConnectionError) as e:
          logger.exception("")
          if test_result is not None:
            status_dict = getattr(e, "status_dict", None) or {
              'stderr': "%s: %s" % (e.__class__.__name__, e)}
            test_result.reportFailure(
              command=status_dict.get('command'),
              stdout=status_dict.get('stdout'),
              stderr=status_dict.get('stderr'),
            )
          continue
        except ValueError as e:
          # This could at least happens if runTestSuite is not found
          logger.exception("")
          if node_test_suite is not None:
            node_test_suite.retry_software_count += 1
          if test_result is not None:
            test_result.reportFailure(
              command='', stdout='',
              stderr="ValueError was raised : %s" % (e,),
            )
        except CancellationError:
          logger.exception("")
          self.process_manager.under_cancellation = False
          node_test_suite.retry = True
          continue
        self.cleanUp()
        sleep_time = 120 - (time.time() - begin)
        if sleep_time > 0:
          logger.info("End of processing, going to sleep %s", sleep_time)
          time.sleep(sleep_time)
    except Exception:
      logger.exception("")
    except:
      logger.exception("")
      raise
    finally:
      # Nice way to kill *everything* generated by run process -- process
      # groups working only in POSIX compilant systems
      # Exceptions are swallowed during cleanup phase
      logger.info("GENERAL EXCEPTION, QUITING")
      self.cleanUp()
      logger.info("GENERAL EXCEPTION, QUITING, cleanup finished")


import unittest
from unittest import TestCase
from contextlib import contextmanager

from xml_marshaller import xml_marshaller
from erp5.util.testnode import logger
from erp5.util.testnode.testnode import TestNode, test_type_registry
from erp5.util.testnode.NodeTestSuite import SlapOSInstance, NodeTestSuite
from erp5.util.testnode.ProcessManager import ProcessManager, SubprocessError
from erp5.util.testnode.Updater import Updater

from erp5.util.testnode.SlapOSMasterCommunicator import SlapOSMasterCommunicator
from erp5.util.testnode.SlapOSControler import SlapOSControler
from erp5.util.testnode.SlapOSControler import createFolder

from erp5.util.taskdistribution import TaskDistributor
from erp5.util.taskdistribution import TestResultProxy
import argparse
import logging
import os
import shutil
import subprocess
import sys
import tempfile
import json
import time
import re
from six.moves.configparser import ConfigParser
try:
  from unittest import mock
except ImportError:
  # BBB python2
  import mock

@contextmanager
def dummySuiteLog(_):
  yield

class ERP5TestNode(TestCase):

  def setUp(self):
    self._temp_dir = tempfile.mkdtemp()
    self.working_directory = os.path.join(self._temp_dir, 'testnode')
    self.slapos_directory = os.path.join(self._temp_dir, 'slapos')
    self.software_directory = os.path.join(self._temp_dir, 'software_directory')
    self.test_suite_directory = os.path.join(self._temp_dir,'test_suite')
    self.environment = os.path.join(self._temp_dir,'environment')
    self.log_directory = os.path.join(self._temp_dir,'var/log/testnode')
    self.log_file = os.path.join(self.log_directory,'test.log')
    self.remote_repository0 = os.path.join(self._temp_dir, 'rep0')
    self.remote_repository1 = os.path.join(self._temp_dir, 'rep1')
    self.remote_repository2 = os.path.join(self._temp_dir, 'rep2')
    self.remote_repository_broken = os.path.join(self._temp_dir, 'broken')
    self.system_temp_folder = os.path.join(self._temp_dir,'tmp')
    os.mkdir(self.working_directory)
    os.mkdir(self.slapos_directory)
    os.mkdir(self.test_suite_directory)
    os.mkdir(self.software_directory)
    os.mkdir(self.environment)
    os.mkdir(self.system_temp_folder)
    os.makedirs(self.log_directory)
    os.close(os.open(self.log_file,os.O_CREAT))
    os.mkdir(self.remote_repository0)
    os.mkdir(self.remote_repository1)
    os.mkdir(self.remote_repository2)

  def tearDown(self):
    shutil.rmtree(self._temp_dir, True)

  def getTestNode(self):
    config = {}
    config["git_binary"] = "git"
    config["master_url"] = 'http://example.org/'
    config["proxy_host"] = ''
    config["proxy_port"] = ''
    config["slapos_directory"] = self.slapos_directory
    config["working_directory"] = self.working_directory
    config["software_directory"] = self.software_directory
    config["node_quantity"] = '3'
    config["test_suite_directory"] = self.test_suite_directory
    config["environment"] = self.environment
    config["log_directory"] = self.log_directory
    config["log_file"] = self.log_file
    config["test_suite_master_url"] = None
    config["hateoas_slapos_master_url"] = None
    config["test_node_title"] = "Foo-Test-Node"
    config["system_temp_folder"] = self.system_temp_folder
    config["computer_id"] = "COMP-TEST"
    config["server_url"] = "http://foo.bar"
    config["httpd_ip"] = "ff:ff:ff:ff:ff:ff:ff:ff"
    config["httpd_software_access_port"] = "9080"
    config["frontend_url"] = "http://frontend/"
    config["log_frontend_url"] = "http://log-frontend/"
    config["software_list"] = ["foo", "bar"]
    config["partition_reference"] = "part"
    config["ipv4_address"] = "1.2.3.4"
    config["ipv6_address"] = "::1"
    config["slapos_binary"] = "/opt/slapgrid/HASH/bin/slapos"
    config["srv_directory"] = "srv_directory"
    config["shared_part_list"] = "/not/exists\n /not/exists_either"

    testnode = TestNode(config)
    # By default, keep suite logs to stdout for easier debugging
    # (stdout/stderr are automatically reported to ERP5).
    # This is unset by test methods that check normal suite logging.
    testnode.suiteLog = dummySuiteLog
    return testnode

  def getTestSuiteData(self, add_third_repository=False,
                       add_broken_repository=False, reference="foo"):
    data = [{
       "test_suite": "Foo",
       "project_title": reference,
       "test_suite_title": "Foo-Test",
       "test_suite_reference": reference,
       "vcs_repository_list": [
            {'url': self.remote_repository0,
             'profile_path': 'software.cfg',
             'branch': 'master'},
            {'url': self.remote_repository1,
             'buildout_section_id': 'rep1',
             'branch': 'master'}]}]
    if add_third_repository:
      # add a third repository
      # insert in position zero since we already had bug when the profile_path
      # was defined in non-zero position when generating the profile
      data[0]['vcs_repository_list'].insert(0,
            {'url': self.remote_repository2,
             'buildout_section_id': 'rep2',
             'branch': 'foo'})
    if add_broken_repository:
      data[0]['vcs_repository_list'].append(
            {'url': self.remote_repository_broken,
             'buildout_section_id': 'rep2',
             'branch': 'foo'})
    return data

  def updateNodeTestSuiteData(self, node_test_suite,
                              add_third_repository=False,
                              add_broken_repository=False):
    """
    Update from zero/Regenerate the testsuite
    """
    node_test_suite.edit(
       **self.getTestSuiteData(add_third_repository=add_third_repository,
                               add_broken_repository=add_broken_repository)[0])

  def getCaller(self, **kw):
    class Caller(object):

      def __init__(self, **kw):
        self.__dict__.update(**kw)

      def __call__(self, command):
        return subprocess.check_output(command, universal_newlines=True,
                                       **self.__dict__)
    return Caller(**kw)

  def generateTestRepositoryList(self, add_third_repository=False):
    commit_dict = {}
    repository_list = [self.remote_repository0, self.remote_repository1]
    if add_third_repository:
      repository_list.append(self.remote_repository2)
    for i, repository_path in enumerate(repository_list):
      call = self.getCaller(cwd=repository_path)
      call("git init".split())
      git_config = open(os.path.join(repository_path, '.git', 'config'), 'a')
      git_config.write("""
[user]
  name = a b
  email = a@b.c
""")
      git_config.close()
      call("touch first_file".split())
      call("git add first_file".split())
      call("git commit -v -m first_commit".split())
      my_file = open(os.path.join(repository_path, 'first_file'), 'w')
      my_file.write("initial_content%i" % i)
      my_file.close()
      call("git commit -av -m next_commit".split())
      output = call(['git', 'log', '--format=%H %s'])
      output = output.strip()
      output_line_list = output.split("\n")
      self.assertEqual(2, len(output_line_list))
      expected_commit_subject_list = ["next_commit", "first_commit"]
      commit_subject_list = [x.split()[1] for x in output_line_list]
      self.assertEqual(expected_commit_subject_list, commit_subject_list)
      commit_dict['rep%i' % i] = [x.split() for x in output_line_list]
      if repository_path == self.remote_repository2:
        output = call('git checkout master -b foo'.split())
    # commit_dict looks like
    # {'rep1': [['6669613db7239c0b7f6e1fdb82af6f583dcb3a94', 'next_commit'],
    #           ['4f1d14de1b04b4f878a442ee859791fa337bcf85', 'first_commit']],
    #  'rep0': [['fb2a61882148d705fd10ecd87278b458a59920a9', 'next_commit'],
    #           ['4f1d14de1b04b4f878a442ee859791fa337bcf85', 'first_commit']]}
    return commit_dict

  def test_01_getDelNodeTestSuite(self):
    """
    We should be able to get/delete NodeTestSuite objects inside test_node
    """
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.assertEqual(0, node_test_suite.retry_software_count)
    node_test_suite.retry_software_count = 2
    self.assertIs(node_test_suite, test_node.getNodeTestSuite('foo'))
    self.assertEqual(2, node_test_suite.retry_software_count)
    del test_node.node_test_suite_dict['foo']
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.assertEqual(0, node_test_suite.retry_software_count)

  def test_02_NodeTestSuiteWorkingDirectory(self):
    """
    Make sure we extend the working path with the node_test_suite reference
    """
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.assertEqual("%s/foo" % self.working_directory,
                      node_test_suite.working_directory)
    self.assertEqual("%s/foo/test_suite" % self.working_directory,
                      node_test_suite.test_suite_directory)

  def test_03_NodeTestSuiteCheckDataAfterEdit(self):
    """
    When a NodeTestSuite instance is edited, the method _checkData
    analyse properties and add new ones
    """
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    self.assertEqual(2, len(node_test_suite.vcs_repository_list))
    repository_path_list = []
    for vcs_repository in node_test_suite.vcs_repository_list:
      repository_path_list.append(vcs_repository['repository_path'])
    expected_list = ["%s/rep0" % node_test_suite.working_directory,
                     "%s/rep1" % node_test_suite.working_directory]
    self.assertEqual(expected_list, repository_path_list)

  def test_04_constructProfile(self, my_test_type='UnitTest'):
    """
    Check if the software profile is correctly generated
    """
    test_node = self.getTestNode()
    test_node.test_suite_portal = TaskDistributor
    test_node.test_suite_portal.getTestNode = TaskDistributor.getTestType
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite, add_third_repository=True)
    node_test_suite.revision_list = (('rep1', (1234, 'azerty')),
                                     ('rep2', (3456, 'qwerty')))
    test_node.constructProfile(node_test_suite,my_test_type)
    self.assertEqual("%s/software.cfg" % (node_test_suite.working_directory,),
                      node_test_suite.custom_profile_path)
    profile = open(node_test_suite.custom_profile_path, 'r')
    if my_test_type=='UnitTest':
      expected_profile = """\
[buildout]
extends = %(temp_dir)s/testnode/foo/rep0/software.cfg

[rep1]
repository = %(temp_dir)s/testnode/foo/rep1
branch = master
revision =
develop = false
shared = true

[rep2]
repository = %(temp_dir)s/testnode/foo/rep2
branch = foo
revision =
develop = false
shared = true
""" % {'temp_dir': self._temp_dir}
    else:
      revision1 = "azerty"
      revision2 = "qwerty"
      expected_profile = """\
[buildout]
extends = %(temp_dir)s/testnode/foo/rep0/software.cfg

[rep1]
repository = %(temp_dir)s/rep1
revision = %(revision1)s
ignore-ssl-certificate = true
develop = false
shared = true

[rep2]
repository = %(temp_dir)s/rep2
revision = %(revision2)s
ignore-ssl-certificate = true
develop = false
shared = true
""" % {'temp_dir': self._temp_dir, 'revision1': revision1, 'revision2': revision2}
    self.assertEqual(expected_profile, profile.read())
    profile.close()

  def getAndUpdateFullRevisionList(self, test_node, node_test_suite):
    if test_node.updateRevisionList(node_test_suite):
      return node_test_suite.revision.split(',')

  def test_05_updateRevisionList(self):
    """
    Check if we clone correctly repositories and get right revisions
    """
    commit_dict = self.generateTestRepositoryList()
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(2, len(rev_list))
    self.assertEqual(rev_list[0], 'rep0=2-%s' % commit_dict['rep0'][0][0])
    self.assertEqual(rev_list[1], 'rep1=2-%s' % commit_dict['rep1'][0][0])
    my_file = open(os.path.join(self.remote_repository1, 'first_file'), 'w')
    my_file.write("next_content")
    my_file.close()
    call = self.getCaller(cwd=self.remote_repository1)
    call("git commit -av -m new_commit".split())
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertTrue(rev_list[0].startswith('rep0=2-'))
    self.assertTrue(rev_list[1].startswith('rep1=3-'))
    self.assertEqual(2, len(node_test_suite.vcs_repository_list))
    for vcs_repository in node_test_suite.vcs_repository_list:
      self.assertTrue(os.path.exists(vcs_repository['repository_path']))

  def test_05b_changeRepositoryBranch(self, my_test_type='UnitTest'):
    """
    It could happen that the branch is changed for a repository. Testnode must
    be able to reset correctly the branch
    """
    self.generateTestRepositoryList(add_third_repository=True)
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite, add_third_repository=True)
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(3, len(rev_list))
    self.assertEqual(3, len(node_test_suite.vcs_repository_list))
    rep2_clone_path = next(
        x['repository_path']
        for x in node_test_suite.vcs_repository_list
        if x['repository_path'].endswith("rep2"))
    call = self.getCaller(cwd=rep2_clone_path)
    output = call("git branch".split()).strip()
    self.assertTrue("* foo" in output.split('\n'))
    vcs_repository_info = node_test_suite.vcs_repository_list[0]
    self.assertEqual(vcs_repository_info['repository_id'], 'rep2')
    self.assertEqual(vcs_repository_info['branch'], 'foo')
    # change it to master
    vcs_repository_info['branch'] = 'master'
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    output = call("git branch".split()).strip()
    print(output)
    self.assertTrue("* master" in output.split('\n'))
    # Add a third branch on remote, make sure we could switch to it
    remote_call = self.getCaller(cwd=self.remote_repository2)
    output = remote_call('git checkout master -b bar'.split())
    vcs_repository_info['branch'] = 'bar'
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    output = call("git branch".split()).strip()
    self.assertTrue("* bar" in output.split('\n'))
    # Add a fourth branch on remote, make sure we could switch to it
    # this time the branch name is a substring of previous one (we had
    # failure is such case at some point)
    remote_call = self.getCaller(cwd=self.remote_repository2)
    output = remote_call('git checkout master -b ba'.split())
    vcs_repository_info['branch'] = 'ba'
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    output = call("git branch".split()).strip()
    self.assertTrue("* ba" in output.split('\n'))

  def test_05c_changeRepositoryUrl(self):
    """
    It could happen that the url is changed for a repository (new place, or
    change of username and password). testnode must be able to erase and clone
    again the repository
    """
    self.generateTestRepositoryList(add_third_repository=True)
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(2, len(rev_list))
    self.assertEqual(2, len(node_test_suite.vcs_repository_list))
    # patch deleteRepository to make sure it will be called once for the wrong
    # repos, and not for the repos which has not changed
    deleted_repository_path_list = []
    original_deleteRepository = Updater.deleteRepository
    try:
      def deleteRepository(self):
        deleted_repository_path_list.append(self.repository_path)
        original_deleteRepository(self)
      Updater.deleteRepository = deleteRepository
      # change the url of the first repository
      vcs_repository_info = node_test_suite.vcs_repository_list[0]
      vcs_repository_info["url"] = self.remote_repository2
      rep0_clone_path = next(
          x['repository_path']
          for x in node_test_suite.vcs_repository_list
          if x['repository_path'].endswith("rep0"))
      call = self.getCaller(cwd=rep0_clone_path)
      self.assertEqual(call("git config --get remote.origin.url".split()).strip(),
                        self.remote_repository0)
      rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
      self.assertEqual(call("git config --get remote.origin.url".split()).strip(),
                        self.remote_repository2)
      self.assertEqual([rep0_clone_path], deleted_repository_path_list)
    finally:
      Updater.deleteRepository = original_deleteRepository

  def test_05d_LocalModificationOnRepository(self):
    """
    It could happen that there is local modification to to either bug of
    git or any manual operation.
    Testnode must be able reset the repository to make sure we have no failures
    when updating repository
    """
    self.generateTestRepositoryList(add_third_repository=True)
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(2, len(rev_list))
    self.assertEqual(2, len(node_test_suite.vcs_repository_list))
    rep0_clone_path = next(
        x['repository_path']
        for x in node_test_suite.vcs_repository_list
        if x['repository_path'].endswith("rep0"))
    my_file = open(os.path.join(rep0_clone_path, 'first_file'), 'w')
    my_file.write("next_content")
    my_file.close()
    # make sure code still works
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(2, len(rev_list))
    self.assertEqual(2, len(node_test_suite.vcs_repository_list))
    # and check local change was resetted
    my_file = open(os.path.join(rep0_clone_path, 'first_file'), 'r')
    self.assertEqual("initial_content0", my_file.read())
    my_file.close()

  def test_05e_IgnoringIncorrectRepository(self):
    """
    If someone add a test suite with a bad url for git repository (or wrong
    crendentials), the testnode should not block forever and should work on
    other test suites. This method should be able to run
    """
    self.generateTestRepositoryList()
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite, add_broken_repository=True)
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(None, rev_list)

  def test_05f_ResetCorruptedRepository(self):
    """
    It could happens that repository are corrupted, like when server is restarted
    in the middle of a git checkout. So make sure testnode remove repositories
    that are not working any longer.

    In this test, we cause an artificial corruption by truncating git
    repository's index, then we call updateRevisionList to check that it can
    recover from this corruption.
    """
    self.generateTestRepositoryList()
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertTrue(rev_list is not None)
    rep0_clone_path = next(
        x['repository_path']
        for x in node_test_suite.vcs_repository_list
        if x['repository_path'].endswith("rep0"))
    # simulate a data corruption on rep0's index
    with open(os.path.join(rep0_clone_path, '.git', 'index'), 'wb') as index_file:
      index_file.write(
        # magic
        b'\x44\x49\x52\x43'
        # version
        b'\x00\x00\x00\x02'
        # number of entries (1)
        b'\x00\x00\x00\x01'
        # nothing more, the index entry is missing from this corrupted file.
      )

    # we get rev list with corrupted repository, we get None, but in the same
    # time the bad repository is deleted
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(None, rev_list)
    # So next update should go fine
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertTrue(rev_list is not None)

  def test_update_revision_with_head_at_merge(self):
    """Test update revision when the head of the branch is a merge commit.
    """
    self.generateTestRepositoryList(add_third_repository=True)
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite, add_third_repository=True)
    rev_list = self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    self.assertEqual(3, len(rev_list))
    self.assertEqual(3, len(node_test_suite.vcs_repository_list))
    rep2_remote_path = next(
        x['url']
        for x in node_test_suite.vcs_repository_list
        if x['url'].endswith("rep2"))
    call = self.getCaller(cwd=rep2_remote_path)

    # make a branch and merge it, to have a topology like this:
    #
    # o1 -- o2 -- o4
    #        \   /
    #          o3
    call("git checkout -b test_branch".split())
    with open(os.path.join(rep2_remote_path, 'test_file'), 'w') as test_file:
      test_file.write("test file !")
    call("git add test_file".split())
    call("git commit -m added".split())
    call("git checkout master".split())
    call("git merge --no-ff test_branch".split())
    expected_revision = call("git rev-parse HEAD".split()).strip()
    # Including the merge, we have 4 commits on this branch, because master is at o4
    self.assertEqual(
        '4',
        call('git rev-list --topo-order --count HEAD'.split()).strip())

    vcs_repository_info = node_test_suite.vcs_repository_list[0]
    vcs_repository_info['branch'] = 'master'
    rev_list = [
        rev.split('=') for rev
        in self.getAndUpdateFullRevisionList(test_node, node_test_suite)]
    self.assertEqual(
      ["4-" + expected_revision],
      [revision for (repository, revision) in rev_list if repository == 'rep2'])

  def test_06_checkRevision(self):
    """
    Check if we are able to restore older commit hash if master decide so
    """
    commit_dict = self.generateTestRepositoryList()
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    self.getAndUpdateFullRevisionList(test_node, node_test_suite)
    def getRepInfo(count=0, hash=0):
      assert count or hash
      info_list = []
      for vcs_repository in node_test_suite.vcs_repository_list:
        call = self.getCaller(cwd=vcs_repository['repository_path'])
        if count:
          info_list.append(
            call("git rev-list --topo-order --count HEAD".split()).strip())
        if hash:
          info_list.append(
            call("git log -n1 --format=%H".split()).strip())
      return info_list
    self.assertEqual(['2', '2'], getRepInfo(count=1))
    self.assertEqual([commit_dict['rep0'][0][0],commit_dict['rep1'][0][0]],
                      getRepInfo(hash=1))
    class TestResult(object):
      revision = NodeTestSuite.revision
    test_result = TestResult()
    # for test result to be one commit late for rep1 to force testnode to
    # reset tree to older version
    test_result.revision_list = (('rep0', (2, commit_dict['rep0'][0][0])),
                                 ('rep1', (1, commit_dict['rep1'][1][0])))
    test_node.checkRevision(test_result, node_test_suite)
    self.assertEqual(['2', '1'], getRepInfo(count=1))
    self.assertEqual([commit_dict['rep0'][0][0],commit_dict['rep1'][1][0]],
                      getRepInfo(hash=1))

  def test_07_checkExistingTestSuite(self):
    test_node = self.getTestNode()
    test_suite_data = self.getTestSuiteData(add_third_repository=True)
    self.assertEqual([], os.listdir(self.working_directory))
    test_node.purgeOldTestSuite(test_suite_data)
    self.assertEqual([], os.listdir(self.working_directory))
    os.mkdir(os.path.join(self.working_directory, 'foo'))
    self.assertEqual(['foo'], os.listdir(self.working_directory))
    test_node.purgeOldTestSuite(test_suite_data)
    self.assertEqual(['foo'], os.listdir(self.working_directory))
    os.mkdir(os.path.join(self.working_directory, 'bar'))
    self.assertEqual(set(['bar','foo']),
                      set(os.listdir(self.working_directory)))
    test_node.purgeOldTestSuite(test_suite_data)
    self.assertEqual(['foo'], os.listdir(self.working_directory))

  def test_purgeOldTestSuiteChmodNonWriteable(self):
    """Old test suites can be deleted even when some files/directories have
    been chmod'd to make non-writeable  """
    test_node = self.getTestNode()
    test_suite_data = self.getTestSuiteData(add_third_repository=True)
    os.mkdir(os.path.join(self.working_directory, 'bar'))
    non_writable_file = open(os.path.join(self.working_directory, 'bar', 'non-writable-file'), 'w')
    non_writable_file.close()

    os.chmod(os.path.join(self.working_directory, 'bar', 'non-writable-file'), 0o400) # -r--------
    os.chmod(os.path.join(self.working_directory, 'bar'), 0o500) # dr-x------

    test_node.purgeOldTestSuite(test_suite_data) # should not fail
    self.assertEqual([], os.listdir(self.working_directory))

  def test_purgeOldTestSuiteChmodNonWriteableNonReadable(self):
    """Old test suites can be deleted even when some files/directories have
    been chmod'd to make them non readable and non writeable. """
    test_node = self.getTestNode()
    test_suite_data = self.getTestSuiteData(add_third_repository=True)
    os.mkdir(os.path.join(self.working_directory, 'bar'))
    non_writable_file = open(os.path.join(self.working_directory, 'bar', 'non-writable-file'), 'w')
    non_writable_file.close()

    os.chmod(os.path.join(self.working_directory, 'bar', 'non-writable-file'), 0o000) # ----------
    os.chmod(os.path.join(self.working_directory, 'bar'), 0o000) # d---------

    test_node.purgeOldTestSuite(test_suite_data) # should not fail
    self.assertEqual([], os.listdir(self.working_directory))

  def test_distributor_url_obfuscated_in_logs(self):
    test_node = self.getTestNode()
    del test_node.suiteLog

    runner = test_type_registry['UnitTest'](test_node)
    node_test_suite = test_node.getNodeTestSuite('foo')
    self.updateNodeTestSuiteData(node_test_suite)
    node_test_suite.revision_list = ('dummy', (0, '')),

    path = runner.getInstanceRoot(node_test_suite) + '/a/bin/runTestSuite'
    os.makedirs(os.path.dirname(path))
    with tempfile.NamedTemporaryFile(mode='r') as temp_file:
      with open(path, 'w') as f:
        f.write("""#!/bin/sh
        echo runTestSuite called with: "$@" | tee {}
        """.format(temp_file.name))
      os.chmod(path, 0o700)

      with test_node.suiteLog(node_test_suite),\
          mock.patch.object(logger, '_log') as http_logger,\
          mock.patch.object(logging.getLogger(), '_log') as root_logger,\
          mock.patch.object(logging.getLogger(), 'isEnabledFor', return_value=True):
        runner.runTestSuite(node_test_suite, "https://user:secret@example.com/portal_distributions")
      http_logger.assert_called()
      logged_messages = [str(logged) for logged in http_logger.mock_calls]

      # distributor URL is not displayed in logs
      self.assertFalse([m for m in logged_messages if "https://user:secret@example.com/portal_distributions" in m])
      # instead we display obfuscated $DISTRIBUTOR_URL
      self.assertTrue([m for m in logged_messages if "--master_url $DISTRIBUTOR_URL" in m])

      # but the runTestSuite program is called with the actual distributor URL
      self.assertEqual(
        'runTestSuite called with: --master_url https://user:secret@example.com/portal_distributions'
        ' --revision dummy=0- --test_node_title Foo-Test-Node --test_suite Foo --test_suite_title Foo-Test',
        temp_file.read().strip())

      # The root logger (which logs only in var/log/erp5testnode.log) contain the non obfuscated
      # command, because it's sometimes useful for debugging.
      root_logger.assert_called()
      self.assertIn(
          '--master_url https://user:secret@example.com/portal_distributions',
          str(root_logger.mock_calls[-1]))


  def test_09_runTestSuite(self, my_test_type='UnitTest'):
    """
    Check parameters passed to runTestSuite
    Also make sure that optional parameters are passed when needed
    """
    call_parameter_list = []
    parser = argparse.ArgumentParser()
    parser.add_argument('--foo', type=int)
    parser.add_argument('--hello_world', help='Hello world!')
    def spawn(*args, **kw):
      if args[1] == '--help':
        return {'stdout': parser.format_help().encode()}
      call_parameter_list.append(args)

    test_node = self.getTestNode()
    del test_node.suiteLog
    with mock.patch.object(test_node.process_manager, 'spawn', side_effect=spawn) as spawn_mock:
      runner = test_type_registry[my_test_type](test_node)
      # Create and initialise/regenerate a nodetestsuite
      node_test_suite = test_node.getNodeTestSuite('foo')
      self.updateNodeTestSuiteData(node_test_suite)
      node_test_suite.revision_list = ('dummy', (0, '')),

      path = runner.getInstanceRoot(node_test_suite) + '/a/bin/runTestSuite'
      os.makedirs(os.path.dirname(path))
      os.close(os.open(path, os.O_CREAT))

      expected_parameter_list = [path,
        '--master_url', 'http://foo.bar',
        '--revision', 'dummy=0-',
        '--test_node_title', 'Foo-Test-Node',
        '--test_suite', 'Foo',
        '--test_suite_title', 'Foo-Test',
      ]
      def checkRunTestSuiteParameters():
        with test_node.suiteLog(node_test_suite) as suite_log:
          runner.runTestSuite(node_test_suite, "http://foo.bar")
        self.assertEqual(list(call_parameter_list.pop()), expected_parameter_list)
        self.assertFalse(call_parameter_list)
        self.assertEqual(
            spawn_mock.call_args[-1]['SLAPOS_TEST_SHARED_PART_LIST'],
            "/not/exists:/not/exists_either:%s/shared" % node_test_suite.working_directory)
        self.assertEqual(
            spawn_mock.call_args[-1]['SLAPOS_TEST_LOG_DIRECTORY'],
            os.path.join(test_node.config['log_directory'], suite_log))

      checkRunTestSuiteParameters()

      def part(path): # in "bar" SR
        path = test_node.config['slapos_directory'] \
          + '/soft/37b51d194a7513e45b56f6524f2d51f2/parts/' + path
        os.makedirs(os.path.dirname(path))
        os.close(os.open(path, os.O_CREAT))
        return path
      for option in (
          ('--firefox_bin', part('firefox/firefox-slapos')),
          ('--frontend_url', 'http://frontend/'),
          ('--node_quantity', '3'),
          ('--xvfb_bin', part('xserver/bin/Xvfb')),
          ('--shared_part_list', "/not/exists:/not/exists_either:%s/shared" % node_test_suite.working_directory),
        ):
        parser.add_argument(option[0])
        expected_parameter_list += option
        checkRunTestSuiteParameters()

  def test_10_prepareSlapOS(self, my_test_type='UnitTest'):
    test_node = self.getTestNode()
    test_node_slapos = SlapOSInstance(self.slapos_directory)
    runner = test_type_registry[my_test_type](test_node)
    node_test_suite = test_node.getNodeTestSuite('foo')

    with mock.patch(
          'erp5.util.testnode.SlapOSControler.SlapOSControler.runSoftwareRelease',
          return_value={"status_code": 0}
        ) as runSoftwareRelease,\
        mock.patch(
          'erp5.util.testnode.SlapOSControler.SlapOSControler.runComputerPartition',
          return_value={"status_code": 0}
        )  as runComputerPartition,\
        mock.patch('erp5.util.testnode.SlapOSControler.slapos.slap'),\
        mock.patch('subprocess.Popen'):

      runner.prepareSlapOSForTestNode(test_node_slapos)
      self.assertEqual(1, runSoftwareRelease.call_count)
      self.assertEqual(0, runComputerPartition.call_count)

    with mock.patch(
          'erp5.util.testnode.SlapOSControler.SlapOSControler.runSoftwareRelease',
          return_value={"status_code": 0}
        ) as runSoftwareRelease,\
        mock.patch(
          'erp5.util.testnode.SlapOSControler.SlapOSControler.runComputerPartition',
          return_value={"status_code": 0}
        )  as runComputerPartition,\
        mock.patch('erp5.util.testnode.SlapOSControler.slapos.slap'),\
        mock.patch('subprocess.Popen'):

      runner.prepareSlapOSForTestSuite(node_test_suite)
      self.assertEqual(1, runSoftwareRelease.call_count)
      self.assertEqual(1, runComputerPartition.call_count)

    # test node slapos slapos uses the shared parts defined in config
    cfg_parser = ConfigParser()
    with open(os.path.join(test_node_slapos.working_directory, 'slapos.cfg')) as f:
      cfg_parser.readfp(f)
    self.assertEqual(
        '/not/exists\n/not/exists_either',
        cfg_parser.get('slapos', 'shared_part_list'))

    # test suite slapos uses the shared parts from the config, plus
    # a "local" folder for used as shared when installing tested
    # softwares.
    cfg_parser = ConfigParser()
    with open(os.path.join(node_test_suite.working_directory, 'slapos.cfg')) as f:
      cfg_parser.readfp(f)
    self.assertEqual(
        '/not/exists\n/not/exists_either\n%s/shared' % node_test_suite.working_directory,
        cfg_parser.get('slapos', 'shared_part_list'))

    # If running software has status_code 1 we have an error
    with mock.patch(
          'erp5.util.testnode.SlapOSControler.SlapOSControler.runSoftwareRelease',
          return_value={"status_code": 1}
        ) as runSoftwareRelease,\
        mock.patch(
          'erp5.util.testnode.SlapOSControler.SlapOSControler.runComputerPartition',
          return_value={"status_code": 0}
        )  as runComputerPartition,\
        mock.patch('erp5.util.testnode.SlapOSControler.slapos.slap'),\
        mock.patch('subprocess.Popen'):

      self.assertRaises(
          SubprocessError,
          runner.prepareSlapOSForTestSuite,
          node_test_suite)


  def test_11_run(self, my_test_type='UnitTest', grade='master'):
    def doNothing(self, *args, **kw):
      pass
    # Used in case of 'ScalabilityTest'
    def patch_getTestType(self, *args, **kw):
      return my_test_type
    def patch_getSlaposAccountKey(self, *args, **kw):
      return "key"
    def patch_getSlaposAccountCertificate(self, *args, **kw):
      return "Certificate"
    def patch_getSlaposUrl(self, *args, **kw):
      return "http://Foo"
    def patch_getSlaposHateoasUrl(self, *args, **kw):
      return "http://Foo"
    def patch_generateConfiguration(self, *args, **kw):
      return json.dumps({"configuration_list": [{}], "involved_nodes_computer_guid"\
: [], "error_message": "No error.", "launcher_nodes_computer_guid": [], \
"launchable": False, "randomized_path" : "azertyuiop"})
    def patch_isMasterTestnode(self, *args, **kw):
      return (grade == 'master')
    test_self = self
    test_result_path_root = os.path.join(test_self._temp_dir,'test/results')
    os.makedirs(test_result_path_root)
    global counter
    counter = 0
    def patch_startTestSuite(self,node_title,computer_guid='unknown'):
      global counter
      config_list = []
      # Sclalability slave testnode is not directly in charge of testsuites
      if my_test_type == 'ScalabilityTest' and grade == 'slave':
        if counter == 5:
          raise StopIteration
        counter += 1
        return json.dumps([])

      def _checkExistingTestSuite(reference_set):
        test_self.assertEqual(set(reference_set),
                  set(os.listdir(test_node.working_directory)))
        for x in reference_set:
          test_self.assertTrue(os.path.exists(os.path.join(
                             test_node.working_directory,x)),True)
      if counter == 0:
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
        config_list.append(test_self.getTestSuiteData(reference='bar')[0])
      elif counter == 1:
        _checkExistingTestSuite(set(['foo']))
        config_list.append(test_self.getTestSuiteData(reference='bar')[0])
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
      elif counter == 2:
        _checkExistingTestSuite(set(['foo','bar']))
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
        config_list.append(test_self.getTestSuiteData(reference='qux')[0])
      elif counter == 3:
        _checkExistingTestSuite(set(['foo','qux']))
        config_list.append(test_self.getTestSuiteData(reference='foox')[0])
      elif counter == 4:
        _checkExistingTestSuite(set(['foox']))
        config_list.append(test_self.getTestSuiteData(reference='bax')[0])
      elif counter == 5:
        _checkExistingTestSuite(set(['bax']))
        raise StopIteration
      counter += 1
      return json.dumps(config_list)
    def patch_createTestResult(self, revision, test_name_list, node_title,
            allow_restart=False, test_title=None, project_title=None):
      global counter
      # return no test to check if run method will run the next test suite
      if counter != 3 or project_title == 'qux':
        test_result_path = os.path.join(test_result_path_root, test_title)
        return TestResultProxy(self._proxy, self._retry_time,
                logger, test_result_path, node_title, revision)
    def patch_runTestSuite(self, *argv, **kw):
      return {'status_code': 0}
    original_sleep = time.sleep
    time.sleep = doNothing
    self.generateTestRepositoryList()
    RunnerClass = test_type_registry[my_test_type]
    # Patch
    if my_test_type == "ScalabilityTest":
      original_getSlaposAccountKey = TaskDistributor.getSlaposAccountKey
      original_getSlaposAccountCertificate = TaskDistributor.getSlaposAccountCertificate
      original_getSlaposUrl = TaskDistributor.getSlaposUrl
      original_getSlaposHateoasUrl = TaskDistributor.getSlaposHateoasUrl
      original_isMasterTestnode = TaskDistributor.isMasterTestnode
      original_updateInstanceXML = RunnerClass._updateInstanceXML
      original_SlapOSMasterCommunicator__init__ = SlapOSMasterCommunicator.__init__
      TaskDistributor.getSlaposAccountKey = patch_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = patch_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = patch_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = patch_getSlaposHateoasUrl
      TaskDistributor.isMasterTestnode = patch_isMasterTestnode
      RunnerClass._updateInstanceXML = doNothing
      SlapOSMasterCommunicator.__init__ = doNothing
    original_generateConfiguration = TaskDistributor.generateConfiguration
    original_initializeSlapOSControler = SlapOSControler.initializeSlapOSControler

    TaskDistributor.generateConfiguration = patch_generateConfiguration
    original_startTestSuite = TaskDistributor.startTestSuite
    original_subscribeNode = TaskDistributor.subscribeNode
    original_getTestType = TaskDistributor.getTestType
    original_createTestResult = TaskDistributor.createTestResult
    TaskDistributor.startTestSuite = patch_startTestSuite
    TaskDistributor.subscribeNode = doNothing
    TaskDistributor.getTestType = patch_getTestType
    TaskDistributor.createTestResult = patch_createTestResult

    # TestNode
    test_node = self.getTestNode()
    # Modify class UnitTestRunner(or more after) method
    def patch_prepareSlapOS(*args, **kw):
      return {'status_code': 0}
    original_prepareSlapOS = RunnerClass._prepareSlapOS
    original_runTestSuite = RunnerClass.runTestSuite
    RunnerClass._prepareSlapOS = patch_prepareSlapOS
    RunnerClass.runTestSuite = patch_runTestSuite
    SlapOSControler.initializeSlapOSControler = doNothing
    # Inside test_node a runner is created using new UnitTestRunner methods
    test_node.run()
    self.assertEqual(5, counter)
    time.sleep = original_sleep
    # Restore old class methods
    if my_test_type == "ScalabilityTest":
      TaskDistributor.getSlaposAccountKey = original_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = original_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = original_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = original_getSlaposHateoasUrl
      TaskDistributor.isMasterTestnode = original_isMasterTestnode
      RunnerClass._updateInstanceXML = original_updateInstanceXML
      SlapOSMasterCommunicator.__init__ = original_SlapOSMasterCommunicator__init__
    TaskDistributor.generateConfiguration = original_generateConfiguration
    TaskDistributor.startTestSuite = original_startTestSuite
    TaskDistributor.createTestResult = original_createTestResult
    TaskDistributor.subscribeNode = original_subscribeNode
    TaskDistributor.getTestType = original_getTestType
    RunnerClass._prepareSlapOS = original_prepareSlapOS
    RunnerClass.runTestSuite = original_runTestSuite
    SlapOSControler.initializeSlapOSControler = original_initializeSlapOSControler

  def test_12_spawn(self):
    def _checkCorrectStatus(expected_status,*args):
      result = process_manager.spawn(*args)
      self.assertEqual(result['status_code'], expected_status)
    process_manager = ProcessManager(max_timeout=1)
    _checkCorrectStatus(0, *['sleep','0'])
    # We must make sure that if the command is too long that
    # it will be automatically killed
    self.assertRaises(SubprocessError, process_manager.spawn, 'sleep','3')

  def test_13_SlaposControlerResetSoftware(self):
    test_node = self.getTestNode()
    controler = SlapOSControler(self.working_directory, test_node.config)
    os.mkdir(controler.software_root)
    file_name = 'AC_Ra\xc3\xadzertic\xc3\xa1ma'
    non_ascii_file = open(os.path.join(controler.software_root, file_name), 'w')
    non_ascii_file.close()
    self.assertEqual([file_name], os.listdir(controler.software_root))
    controler._resetSoftware()
    self.assertEqual([], os.listdir(controler.software_root))

  def test_14_createFolder(self):
    test_node = self.getTestNode()
    node_test_suite = test_node.getNodeTestSuite('foo')
    folder = node_test_suite.test_suite_directory
    self.assertFalse(os.path.exists(folder))
    createFolder(folder)
    self.assertTrue(os.path.exists(folder))
    to_drop_path = os.path.join(folder, 'drop')
    to_drop = open(to_drop_path, 'w')
    to_drop.close()
    self.assertTrue(os.path.exists(to_drop_path))
    createFolder(folder, clean=True)
    self.assertFalse(os.path.exists(to_drop_path))

  def test_15_suite_log_directory(self, my_test_type='UnitTest', grade='master'):
    def doNothing(self, *args, **kw):
      pass
    # Used in case of 'ScalabilityTest'
    def patch_getTestType(self, *args, **kw):
      return my_test_type
    def patch_getSlaposAccountKey(self, *args, **kw):
      return "key"
    def patch_getSlaposAccountCertificate(self, *args, **kw):
      return "Certificate"
    def patch_getSlaposUrl(self, *args, **kw):
      return "http://Foo"
    def patch_getSlaposHateoasUrl(self, *args, **kw):
      return "http://Foo"
    def patch_generateConfiguration(self, *args, **kw):
      return json.dumps({"configuration_list": [{}], "involved_nodes_computer_guid"\
: [], "error_message": "No error.", "launcher_nodes_computer_guid": [], \
"launchable": False, "randomized_path" : "azertyuiop"})
    def patch_isMasterTestnode(self, *args, **kw):
      return grade == 'master'
    test_self = self
    test_result_path_root = os.path.join(test_self._temp_dir,'test/results')
    os.makedirs(test_result_path_root)
    global counter
    counter = 0
    def patch_startTestSuite(self,node_title,computer_guid='unknown'):
      global counter
      config_list = [test_self.getTestSuiteData(reference='aa')[0],
                     test_self.getTestSuiteData(reference='bb')[0]]
      if counter in (1, 2):
        config_list.reverse()
      elif counter == 3:
        raise StopIteration
      counter += 1
      return json.dumps(config_list)
    def patch_createTestResult(self, revision, test_name_list, node_title,
            allow_restart=False, test_title=None, project_title=None):
      test_result_path = os.path.join(test_result_path_root, test_title)
      return TestResultProxy(self._proxy, self._retry_time,
                logger, test_result_path, node_title, revision)
    def patch_runTestSuite(self,*argv, **kw):
      return {'status_code': 0}
    def checkTestSuite(test_node, reportTaskStatus_mock):
      test_node.node_test_suite_dict
      rand_part_set = set()
      self.assertEqual(2, len(test_node.node_test_suite_dict))
      for suite in test_node.node_test_suite_dict.values():
        self.assertTrue('var/log/testnode/%s' % suite.reference in \
                         suite.suite_log_path,
                         "Incorrect suite log path : %r" % suite.suite_log_path)
        m = re.search('-(.*)/suite.log$', suite.suite_log_path)
        rand_part = m.groups()[0]
        reportTaskStatus_mock.assert_has_calls(
            [mock.call(
                mock.ANY,
                {'command': 'LOG url',
                'stdout': "http://log-frontend/{}-{}".format(suite.reference, rand_part),
                'stderr': ''},
                "Foo-Test-Node")])

        self.assertEqual(len(rand_part), 10)
        self.assertNotIn(rand_part, rand_part_set)
        rand_part_set.add(rand_part)
        with open(suite.suite_log_path) as suite_log:
          self.assertIn("Getting configuration from test suite",
                        suite_log.readline())

    RunnerClass = test_type_registry[my_test_type]
    original_sleep = time.sleep
    time.sleep = doNothing
    self.generateTestRepositoryList()
    if my_test_type == "ScalabilityTest":
      original_getSlaposAccountKey = TaskDistributor.getSlaposAccountKey
      original_getSlaposAccountCertificate = TaskDistributor.getSlaposAccountCertificate
      original_getSlaposUrl = TaskDistributor.getSlaposUrl
      original_getSlaposHateoasUrl = TaskDistributor.getSlaposHateoasUrl
      original_isMasterTestnode = TaskDistributor.isMasterTestnode
      original_supply = SlapOSControler.supply
      original_request = SlapOSControler.request
      original_updateInstanceXML = RunnerClass._updateInstanceXML
      original_SlapOSMasterCommunicator__init__ = SlapOSMasterCommunicator.__init__
      TaskDistributor.getSlaposAccountKey = patch_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = patch_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = patch_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = patch_getSlaposHateoasUrl
      TaskDistributor.isMasterTestnode = patch_isMasterTestnode
      SlapOSControler.supply = doNothing
      SlapOSControler.request = doNothing
      RunnerClass._updateInstanceXML = doNothing
      SlapOSMasterCommunicator.__init__ = doNothing
    original_generateConfiguration = TaskDistributor.generateConfiguration
    original_startTestSuite = TaskDistributor.startTestSuite
    original_subscribeNode = TaskDistributor.subscribeNode
    original_getTestType = TaskDistributor.getTestType
    TaskDistributor.generateConfiguration = patch_generateConfiguration
    TaskDistributor.startTestSuite = patch_startTestSuite
    TaskDistributor.subscribeNode = doNothing
    TaskDistributor.getTestType = patch_getTestType
    original_createTestResult = TaskDistributor.createTestResult
    TaskDistributor.createTestResult = patch_createTestResult
    test_node = self.getTestNode()
    del test_node.suiteLog
    # Change UnitTestRunner class methods
    original_prepareSlapOS = RunnerClass._prepareSlapOS
    original_runTestSuite = RunnerClass.runTestSuite
    original_initializeSlapOSControler = SlapOSControler.initializeSlapOSControler

    if my_test_type == "ScalabilityTest":
      RunnerClass.runTestSuite = patch_runTestSuite
    else:
      RunnerClass.runTestSuite = doNothing

    def patch_prepareSlapOS(*args, **kw):
      return {'status_code': 0}
    RunnerClass._prepareSlapOS = patch_prepareSlapOS
    SlapOSControler.initializeSlapOSControler = doNothing
    with mock.patch('erp5.util.taskdistribution.DummyTaskDistributor.reportTaskStatus') as reportTaskStatus:
      test_node.run()

    self.assertEqual(counter, 3)
    checkTestSuite(test_node, reportTaskStatus)
    time.sleep = original_sleep
    # Restore old class methods
    if my_test_type == "ScalabilityTest":
      TaskDistributor.getSlaposAccountKey = original_getSlaposAccountKey
      TaskDistributor.getSlaposAccountCertificate = original_getSlaposAccountCertificate
      TaskDistributor.getSlaposUrl = original_getSlaposUrl
      TaskDistributor.getSlaposHateoasUrl = original_getSlaposHateoasUrl
      TaskDistributor.isMasterTestnode = original_isMasterTestnode
      SlapOSControler.supply =original_supply
      SlapOSControler.request = original_request
      SlapOSControler.updateInstanceXML = original_updateInstanceXML
      SlapOSMasterCommunicator.__init__ = original_SlapOSMasterCommunicator__init__
    TaskDistributor.generateConfiguration = original_generateConfiguration
    TaskDistributor.startTestSuite = original_startTestSuite
    TaskDistributor.createTestResult = original_createTestResult
    TaskDistributor.subscribeNode = original_subscribeNode
    TaskDistributor.getTestType = original_getTestType
    RunnerClass._prepareSlapOS = original_prepareSlapOS
    RunnerClass.runTestSuite = original_runTestSuite
    SlapOSControler.initializeSlapOSControler = original_initializeSlapOSControler

  def test_16_cleanupLogDirectory(self):
    # Make sure that we are able to cleanup old log folders
    test_node = self.getTestNode()
    def check(file_list):
      log_directory_dir = os.listdir(self.log_directory)
      self.assertTrue(set(file_list).issubset(
           set(log_directory_dir)),
           "%r not contained by %r" % (file_list, log_directory_dir))
    check([])
    os.mkdir(os.path.join(self.log_directory, 'ab-llzje'))
    a_file = open(os.path.join(self.log_directory, 'a_file'), 'w')
    a_file.close()
    check(set(['ab-llzje', 'a_file']))
    # default log file time is 15 days, so nothing is going to be deleted
    test_node._cleanupLog()
    check(set(['ab-llzje', 'a_file']))
    # then we set keep time to 0, folder will be deleted
    test_node.max_log_time = 0
    test_node._cleanupLog()
    check(set(['a_file']))

  def test_17_cleanupTempDirectory(self):
    # Make sure that we are able to cleanup old temp folders
    test_node = self.getTestNode()
    temp_directory = self.system_temp_folder
    def check(file_list):
      directory_dir = os.listdir(temp_directory)
      self.assertTrue(set(file_list).issubset(
           set(directory_dir)),
           "%r not contained by %r" % (file_list, directory_dir))
    check([])
    os.mkdir(os.path.join(temp_directory, 'buildoutA'))
    os.mkdir(os.path.join(temp_directory, 'something')) # this will be kept, as it's not a buildout tempfile
    os.mkdir(os.path.join(temp_directory, 'tmpC'))
    os.mkdir(os.path.join(temp_directory, 'tmp-cannot-delete'), 0o000)

    check(set(['buildoutA', 'something', 'tmpC', 'tmp-cannot-delete']))
    # default log file time is 15 days, so nothing is going to be deleted
    test_node._cleanupTemporaryFiles()
    check(set(['buildoutA', 'something', 'tmpC', 'tmp-cannot-delete']))
    # then we set keep time to 0, folder will be deleted
    test_node.max_temp_time = 0
    test_node._cleanupTemporaryFiles()
    # "something" is kept, as it is not a buildout related tmpfile
    check(set(['something']))
    # other buildout related files are all deleted now.
    self.assertEqual(
        set([]) ,
        set(['buildoutA', 'tmpC', 'tmp-cannot-delete']).intersection(
            set(os.listdir(temp_directory))))

  def test_18_resetSoftwareAfterManyBuildFailures(self, my_test_type='UnitTest'):
    """
    Check that after several building failures that the software is resetted
    """
    initial_initializeSlapOSControler = \
      SlapOSControler.initializeSlapOSControler
    initial_runSoftwareRelease = SlapOSControler.runSoftwareRelease
    test_node = self.getTestNode()
    runner = test_type_registry[my_test_type](test_node)
    node_test_suite = test_node.getNodeTestSuite('foo')
    init_call_kw_list = []
    def initializeSlapOSControler(self, **kw):
      init_call_kw_list.append(kw)
    def runSoftwareRelease(self, *args, **kw):
      return {"status_code": 1}
    SlapOSControler.initializeSlapOSControler = initializeSlapOSControler
    SlapOSControler.runSoftwareRelease = runSoftwareRelease
    def callPrepareSlapOS():
      runner._prepareSlapOS(self.working_directory, node_test_suite,
         create_partition=0)
    def callRaisingPrepareSlapos():
      self.assertRaises(SubprocessError, callPrepareSlapOS)

    self.assertEqual(node_test_suite.retry_software_count, 0)
    for _ in range(11):
      callRaisingPrepareSlapos()
    self.assertEqual(len(init_call_kw_list), 11)
    self.assertEqual(init_call_kw_list[-1]['reset_software'], False)
    self.assertEqual(node_test_suite.retry_software_count, 11)
    callRaisingPrepareSlapos()
    self.assertEqual(init_call_kw_list[-1]['reset_software'], True)
    self.assertEqual(node_test_suite.retry_software_count, 1)
    callRaisingPrepareSlapos()
    self.assertEqual(init_call_kw_list[-1]['reset_software'], False)
    self.assertEqual(node_test_suite.retry_software_count, 2)
    SlapOSControler.initializeSlapOSControler = \
      initial_initializeSlapOSControler
    SlapOSControler.runSoftwareRelease = initial_runSoftwareRelease

  def test_scalability_04_constructProfile(self, my_test_type='ScalabilityTest'):
    self.test_04_constructProfile(my_test_type)
  def test_scalability_05b_changeRepositoryBranch(self, my_test_type='ScalabilityTest'):
    self.test_05b_changeRepositoryBranch(my_test_type)
  def test_scalability_09_runTestSuite(self, my_test_type='ScalabilityTest'):
    # TODO : write own scalability test
    pass
  def test_scalability_10_prepareSlapOS(self, my_test_type='ScalabilityTest'):
    # TODO : write own scalability test
    # This case test may be dispensable on ScalabilityTest case
    # so..
    pass
  @unittest.skip("Not implemented")
  def test_scalability_as_master_11_run(self, my_test_type='ScalabilityTest'):
    self.test_11_run(my_test_type, grade='master')
  # TODO : add a test with master and a launchable testsuite -> patch a lot of methods
  @unittest.skip("Not implemented")
  def test_scalability_as_slave_11_run(self, my_test_type='ScalabilityTest'):
    self.test_11_run(my_test_type, grade='slave')
  @unittest.skip("Not implemented")
  def test_scalability_as_master_15_suite_log_directory(self, my_test_type='ScalabilityTest'):
    self.test_15_suite_log_directory(my_test_type, grade='master')
  @unittest.skip("Not implemented")
  def test_scalability_as_slave_15_suite_log_directory(self, my_test_type='ScalabilityTest'):
    self.test_15_suite_log_directory(my_test_type, grade='slave')
  def test_scalability_18_resetSoftwareAfterManyBuildFailures(self, my_test_type='ScalabilityTest'):
    # TODO : write own scalability test
    pass

  def test_zzzz_scalability_19_xxxx(self):
    # TODO : fill the dummy slapos answer
    # by patching isSoftwareReleaseReady method.
    def patch_createTestResult(self, revision, test_name_list, node_title,
            allow_restart=False, test_title=None, project_title=None):
      test_result_path = os.path.join(test_result_path_root, test_title)
      return TestResultProxy(self._proxy, self._retry_time,
                logger, test_result_path, node_title, revision)
    global startTestSuiteDone
    startTestSuiteDone = False
    def patch_startTestSuite(self,node_title,computer_guid='unknown'):
      config_list = []
      global startTestSuiteDone
      if not startTestSuiteDone:
        startTestSuiteDone = True
        config_list.append(test_self.getTestSuiteData(reference='foo')[0])
        config_list.append(test_self.getTestSuiteData(reference='bar')[0])
      else:
        raise StopIteration
      return json.dumps(config_list)
    def patch_isMasterTestnode(self, *args, **kw):
      return True
    def patch_generateConfiguration(self, *args, **kw):
      return json.dumps({"configuration_list": [{"ok": "ok"}], "involved_nodes_computer_guid"\
: ["COMP1", "COMP2", "COMP3"], "error_message": "No error.", "launcher_nodes_computer_guid": ["COMP1"], \
"launchable": True, "randomized_path" : "azertyuiop"})
    def doNothing(self, *args, **kw):
      pass
    def patch_getSlaposAccountKey(self, *args, **kw):
      return "key"
    def patch_getSlaposAccountCertificate(self, *args, **kw):
      return "Certificate"
    def patch_getSlaposUrl(self, *args, **kw):
      return "http://Foo"
    def patch_getSlaposHateoasUrl(self, *args, **kw):
      return "http://Foo"
    def patch_getTestType(self, *args, **kw):
      return "ScalabilityTest"
    def patch_runTestSuite(self, *args, **kw):
      return {'status_code': 0}
    def patch_generateProfilePasswordAccess(self, *args, **kw):
      return "user", "pass"
    def patch_getDictionaryFromFile(self, *args, **kw):
      return []
    test_self = self
    test_result_path_root = os.path.join(test_self._temp_dir,'test/results')
    os.makedirs(test_result_path_root)
    self.generateTestRepositoryList()
    # Select the good runner to modify
    RunnerClass = test_type_registry['ScalabilityTest']
    # Patch methods
    original_sleep = time.sleep
    original_getSlaposAccountKey = TaskDistributor.getSlaposAccountKey
    original_getSlaposAccountCertificate = TaskDistributor.getSlaposAccountCertificate
    original_getSlaposUrl = TaskDistributor.getSlaposUrl
    original_getSlaposHateoasUrl = TaskDistributor.getSlaposHateoasUrl
    original_generateConfiguration = TaskDistributor.generateConfiguration
    original_isMasterTestnode = TaskDistributor.isMasterTestnode
    original_startTestSuite = TaskDistributor.startTestSuite
    original_subscribeNode = TaskDistributor.subscribeNode
    original_getTestType = TaskDistributor.getTestType
    original_createTestResult = TaskDistributor.createTestResult
    original_prepareSlapOS = RunnerClass._prepareSlapOS
    original_runTestSuite = RunnerClass.runTestSuite
    original_supply = SlapOSControler.supply
    original_request = SlapOSControler.request
    original_updateInstanceXML = SlapOSControler.updateInstanceXML
    original_SlapOSMasterCommunicator__init__ = SlapOSMasterCommunicator.__init__
    original_generateProfilePasswordAccess = RunnerClass.generateProfilePasswordAccess
    original_getDictionaryFromFile = RunnerClass.getDictionaryFromFile
    original_updateDictionaryFile = RunnerClass.updateDictionaryFile
    original__createInstance = RunnerClass._createInstance
    original__waitInstanceCreation = RunnerClass._waitInstanceCreation

    time.sleep = doNothing
    TaskDistributor.getSlaposAccountKey = patch_getSlaposAccountKey
    TaskDistributor.getSlaposAccountCertificate = patch_getSlaposAccountCertificate
    TaskDistributor.getSlaposUrl = patch_getSlaposUrl
    TaskDistributor.getSlaposHateoasUrl = patch_getSlaposHateoasUrl
    TaskDistributor.generateConfiguration = patch_generateConfiguration
    TaskDistributor.isMasterTestnode = patch_isMasterTestnode
    TaskDistributor.startTestSuite = patch_startTestSuite
    TaskDistributor.subscribeNode = doNothing
    TaskDistributor.getTestType = patch_getTestType
    TaskDistributor.createTestResult = patch_createTestResult
    RunnerClass._prepareSlapOS = doNothing
    RunnerClass.runTestSuite = patch_runTestSuite
    SlapOSControler.supply = doNothing
    SlapOSControler.request = doNothing
    SlapOSControler.updateInstanceXML = doNothing
    SlapOSMasterCommunicator.__init__ = doNothing
    RunnerClass.generateProfilePasswordAccess = patch_generateProfilePasswordAccess
    RunnerClass.updateDictionaryFile = doNothing
    RunnerClass.getDictionaryFromFile = patch_getDictionaryFromFile
    RunnerClass._createInstance = doNothing
    RunnerClass._waitInstanceCreation = doNothing
    # Run
    test_node = self.getTestNode()
    test_node.run()
    # Restore methods
    TaskDistributor.getSlaposAccountKey = original_getSlaposAccountKey
    TaskDistributor.getSlaposAccountCertificate = original_getSlaposAccountCertificate
    TaskDistributor.getSlaposUrl = original_getSlaposUrl
    TaskDistributor.getSlaposHateoasUrl = original_getSlaposHateoasUrl
    TaskDistributor.generateConfiguration = original_generateConfiguration
    TaskDistributor.isMasterTestnode = original_isMasterTestnode
    TaskDistributor.startTestSuite = original_startTestSuite
    TaskDistributor.createTestResult = original_createTestResult
    TaskDistributor.subscribeNode = original_subscribeNode
    TaskDistributor.getTestType = original_getTestType
    RunnerClass._prepareSlapOS = original_prepareSlapOS
    RunnerClass.runTestSuite = original_runTestSuite
    SlapOSControler.supply = original_supply
    SlapOSControler.request = original_request
    SlapOSControler.updateInstanceXML = original_updateInstanceXML
    SlapOSMasterCommunicator.__init__ = original_SlapOSMasterCommunicator__init__
    time.sleep =original_sleep
    RunnerClass.generateProfilePasswordAccess = original_generateProfilePasswordAccess
    RunnerClass.getDictionaryFromFile = original_getDictionaryFromFile
    RunnerClass.updateDictionaryFile = original_updateDictionaryFile
    RunnerClass._createInstance = original__createInstance
    RunnerClass._waitInstanceCreation = original__waitInstanceCreation

  def test_initializeSlapOSControler(self):
    test_node = self.getTestNode()
    test_node_slapos = SlapOSInstance(self.slapos_directory)
    runner = test_type_registry['UnitTest'](test_node)

    import slapos.slap
    Computer = mock.create_autospec(slapos.slap.Computer)
    with mock.patch(
          'erp5.util.testnode.SlapOSControler.SlapOSControler.runSoftwareRelease',
          return_value={"status_code": 0}
        ),\
        mock.patch('erp5.util.testnode.SlapOSControler.slapos.slap.slap.registerComputer', return_value=Computer),\
        mock.patch('erp5.util.testnode.SlapOSControler.slapos.slap.slap.initializeConnection'),\
        mock.patch('erp5.util.testnode.SlapOSControler.slapos.slap.slap.registerSupply'),\
        mock.patch('subprocess.Popen'):
      runner.prepareSlapOSForTestNode(test_node_slapos)

    # update configuration is called only once to set all partitions
    self.assertEqual(1, Computer.updateConfiguration.call_count)
    computer_config = xml_marshaller.loads(Computer.updateConfiguration.call_args[0][0])
    self.assertEqual(10, len(computer_config['partition_list']))



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2010 Nexedi SA and Contributors. All Rights Reserved.
#                    Nicolas Delaby <nicolas@nexedi.com>
#
# WARNING: This program as such is intended to be used by professional
# programmers who take the whole responsability of assessing all potential
# consequences resulting from its eventual inadequacies and bugs
# End users who are looking for a ready-to-use solution with commercial
# garantees and support are strongly adviced to contract a Free Software
# Service Company
#
# This program is Free Software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#
##############################################################################
from __future__ import print_function

import os
import shutil
import sys
import re
import shlex
from subprocess import Popen, PIPE, STDOUT
import logging
import tempfile
from datetime import datetime
import threading
import signal
import six

_MARKER = []

class ProcessTimeoutException(Exception):
  pass

def alarm_handler(signum, frame):
  raise ProcessTimeoutException

class HTTPCacheCheckerTestSuite(object):

  URL_CODE = 'url'
  STATUS_CODE = 'status'
  FILE_PATH_CODE = 'path'
  OTHER_CODE = 'other'

  LOG_LEVEL_DICT = {'debug': logging.DEBUG,
                    'info': logging.INFO,
                    'warning': logging.WARNING,
                    'error': logging.ERROR,
                    'critical': logging.CRITICAL}


  url_search_in_wget_regex = re.compile('^--\d{4}.*--\s+(?P<%s>.*)$' %\
                                                                      URL_CODE)
  status_search_in_wget_regex = re.compile('^HTTP request sent, awaiting '\
                                 'response\.\.\. (?P<%s>\d+).+$' % STATUS_CODE)
  file_save_search_regex = re.compile("^Saving to: `(?P<%s>.*)'" %\
                                                                FILE_PATH_CODE)

  x_cache_header_search_regex = re.compile('X-Cache:\s(\S+)\s?$', re.MULTILINE)
  x_varnish_header_search_regex = re.compile('X-Varnish:\s(\d+)', re.MULTILINE)
  generic_header_search_regex = '%s:\s(.*)\s$'

  ACCEPTABLE_STATUS_LIST = ('200', '304', '302',)

  def __init__(self, root_url, working_directory, varnishlog_binary_path,
               wget_binary_path, header_list, email_address, smtp_host,
               debug_level, file_log_path, conditional_header_dict,
               no_header_dict):
    """
      root_url : website to check
      working_directory : where fetched data will be downloaded
      varnishlog_binary_path :  path to varnishlog
      wget_binary_path : path to wget command
      header_list : Key == Header id.
                    value: if equals True means header
                             needs to be present in RESPONSE
                           if this is tuple, the Header value must sastify
                             at least one of the proposed values
      email_address : email address to send result
      smtp_host : smtp host to use
      debug_level : log level of this utility (debug =>very verbose,
                                               info=>normal,
                                               warning=> nothing)
      file_log_path: path to log file
      conditional_header_dict : Key  == Section id (like 'header url='.*/login')
                                value: the configuration lines in this sction
                                       (config format is same as header_list)
      no_header_dict : Key == Section id (like 'no_header url=.*/sitemap')
                       value: = not exsiting headers
    """
    self.root_url = root_url
    self.working_directory = working_directory
    self.varnishlog_binary_path = varnishlog_binary_path
    self.wget_binary_path = wget_binary_path
    self.header_list = header_list
    self.conditional_header_dict = conditional_header_dict
    self.no_header_dict = no_header_dict
    self.email_address = email_address
    self.smtp_host = smtp_host
    level = self.LOG_LEVEL_DICT.get(debug_level, logging.INFO)
    logging.basicConfig(filename=file_log_path, level=level)
    self.report_dict = {}
    self._timeout = 30

  def _initFolder(self):
    """Delete and create workgin directory
    """
    if os.path.isdir(self.working_directory):
      logging.debug('Re-creating folder:%r' % self.working_directory)
      shutil.rmtree(self.working_directory)
      os.mkdir(self.working_directory)
    if not os.path.isdir(self.working_directory):
      logging.debug('Creating folder:%r' % self.working_directory)
      os.mkdir(self.working_directory)

  def _runVarnishLog(self):
    """Run varnishlog and listen comunications
    """
    logging.info('Start varnishlog process')
    command_string = '%s -d' % (self.varnishlog_binary_path)#, filename)
    command_list = shlex.split(command_string)
    process = Popen(command_list, stdin=PIPE, stdout=PIPE)
    return process

  def _readVarnishLog(self, process):
    """Add timeout support
    """
    logging.info('Reading varnishlog with timeout:%r' % self._timeout)
    def _kill_process(pid):
      # support for python 2.5 and under.
      # Shall use process.terminate() for python2.6 and above
      os.kill(pid, signal.SIGTERM)
    watcher = threading.Timer(self._timeout, _kill_process, args=(process.pid,))
    watcher.start()
    varnishlog_data, _ = process.communicate()
    watcher.cancel()
    return varnishlog_data

  def _readVarnishLogAndGetIsBackendTouched(self, varnihslog_data,
                                                     x_varnish_reference_list):
    """Utility to parse such string:

      14 ReqStart     c 127.0.0.1 58470 385643239
      14 RxRequest    c GET
      14 RxURL        c /web/VirtualHostBase/http/www.example.com:80/erp5/web_site_module/example_website/VirtualHostRoot/_vh_/ga/account/
      14 RxProtocol   c HTTP/1.1
      14 RxHeader     c Host: 127.0.0.1:6081
      14 RxHeader     c User-Agent: Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.1.7) Gecko/20100110 Mandriva Linux/1.9.1.7-0.1mdv2010.0 (2010.0) Firefox/3.5.7
      14 RxHeader     c Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
      14 RxHeader     c Accept-Language: fr,fr-fr;q=0.8,en-us;q=0.5,en;q=0.3
      14 RxHeader     c Accept-Encoding: gzip,deflate
      14 RxHeader     c Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7
      14 RxHeader     c Referer: http://www.example.com/
      14 RxHeader     c If-Modified-Since: Thu, 25 Feb 2010 16:22:23 GMT
      14 RxHeader     c Via: 1.1 www.example.com
      14 RxHeader     c X-Forwarded-For: 82.226.226.226
      14 RxHeader     c X-Forwarded-Host: www.example.com
      14 RxHeader     c X-Forwarded-Server: www.example.com
      14 RxHeader     c Connection: Keep-Alive
      14 VCL_call     c recv
      14 VCL_return   c lookup
      14 VCL_call     c hash
      14 VCL_return   c hash
      14 VCL_call     c miss
      14 VCL_return   c fetch
      14 Backend      c 15 default default
      15 TxRequest    b GET
      15 TxURL        b /web/VirtualHostBase/http/www.example.com:80/erp5/web_site_module/example_website/VirtualHostRoot/_vh_/ga/account/
      15 TxProtocol   b HTTP/1.1
      15 TxHeader     b Host: 127.0.0.1:6081
      15 TxHeader     b User-Agent: Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.1.7) Gecko/20100110 Mandriva Linux/1.9.1.7-0.1mdv2010.0 (2010.0) Firefox/3.5.7
      15 TxHeader     b Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
      15 TxHeader     b Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7
      15 TxHeader     b Referer: http://www.example.com/
      15 TxHeader     b Via: 1.1 www.example.com
      15 TxHeader     b X-Forwarded-For: 82.226.226.226
      15 TxHeader     b X-Forwarded-Host: www.example.com
      15 TxHeader     b X-Forwarded-Server: www.example.com
      15 TxHeader     b X-Varnish: 385643239
      15 TxHeader     b X-Forwarded-For: 127.0.0.1
      16 SessionOpen  c 127.0.0.1 58471 :6081
      16 ReqStart     c 127.0.0.1 58471 385643240
      16 RxRequest    c GET
      16 RxURL        c /web/VirtualHostBase/http/www.example.com:80/erp5/web_site_module/example_website/VirtualHostRoot/_vh_/erp5_web_example_layout.css
      16 RxProtocol   c HTTP/1.1
      16 RxHeader     c Host: 127.0.0.1:6081
      [...]

    This script is able to know if the request goes to the backend or not according ID of request
    ie: 385643239
    if we are able to read such string "TxHeader     b X-Varnish: 385643239"
    means varnish send a query to the backend.
    """
    for x_varnish_reference in x_varnish_reference_list:
      backend_is_touched_regex = re.compile('TxHeader\s+b\sX-Varnish:\s%s' %\
                                                        x_varnish_reference[0])
      match_object = backend_is_touched_regex.search(varnihslog_data)
      backend_touched = match_object is not None
      logging.debug('%r %r' % (backend_touched, x_varnish_reference,))
      if backend_touched != x_varnish_reference[1]:
        if backend_touched:
          title = 'Error:Backend touched'
        else:
          title = 'Error Backend not touched'
        message = '%s -> X-Varnish:%r' % (title, x_varnish_reference,)
        self.report_dict.setdefault(x_varnish_reference[2],
                                    []).append(message)

  def _parseWgetLine(self, line):
    """return tuple (code, value)
    code: 'url' means this is requested url
          'status' means we can read status code
          'other' something we can discard
    """
    match_object = self.url_search_in_wget_regex.search(line)
    if match_object is not None:
      return self.URL_CODE, match_object.group(self.URL_CODE)
    match_object = self.status_search_in_wget_regex.search(line)
    if match_object is not None:
      return self.STATUS_CODE, match_object.group(self.STATUS_CODE)
    match_object = self.file_save_search_regex.search(line)
    if match_object is not None:
      return self.FILE_PATH_CODE, match_object.group(self.FILE_PATH_CODE)
    return self.OTHER_CODE, line

  def _runSpider(self):
    """Run wget in working_directory with headers in result
    """
    wget_command_string = '%s -r -nc --retry-connrefused --save-headers %s '\
                                       % (self.wget_binary_path, self.root_url)
    logging.debug('wget command:%r' % wget_command_string)
    wget_argument_list = shlex.split(wget_command_string)
    wget_process = Popen(wget_argument_list, stdin=PIPE,
                         stdout=PIPE, stderr=STDOUT,
                         env={'LANG': 'en_EN'}, # Force output messages
                                                # in English
                         universal_newlines=True,
                         cwd=self.working_directory) # working directory
    stdout, stderr = wget_process.communicate()
    return stdout

  def _getHeaderPolicyList(self, url, fetched_data):
    """
    create the header checking policy list by the url, and the header

    [header_list]
    Last-Modified = True
    Vary = Accept-Language, Cookie, Accept-Encoding
    Cache-Control = max-age=300
    max-age=3600
    [header url=.*/contact_form]
    Last-Modified = True

    [no-header content-type=(image/.*|.*/javascript)]
    Vary = False

    [erp5_extension_list]
    prohibited_folder_name_list = web_page_module
    document_module
    prohibited_file_name_list = WebSection_viewAsWeb
    Base_viewHistory
    list
    """
    def getNoCheckHeaderList(url, fetched_data):
      """
        create no check header list
      """
      pick_content_type = re.compile('^no_header\s*content-type=(.*)')
      pick_url = re.compile('^no_header\s*url=(.*)')
      section_list = self.no_header_dict.keys()
      no_check_header_list = []
      for section in section_list:
        content_type_regex_str_match = pick_content_type.match(section)
        url_regex_str_match = pick_url.match(section)
        if content_type_regex_str_match is not None:
          content_type_regex_str = content_type_regex_str_match.group(1)
          content_type_regex = \
            re.compile('Content-Type:\s%s' % content_type_regex_str,
                       re.MULTILINE | re.IGNORECASE)
          if content_type_regex.search(fetched_data) is not None:
            for header, value in self.no_header_dict[section]:
              no_check_header_list.append(header)
          continue
        if url_regex_str_match is not None:
          url_regex_str = url_regex_str_match.group(1)
          if re.compile(url_regex_str).match(url) is not None:
            for header, value in self.no_header_dict[section]:
              no_check_header_list.append(header)
      return no_check_header_list

    def getConditionalHeaderDict(url, fetched_data):
      """ create header policy by the url and header"""
      conditional_header_dict = {}
      section_list = self.conditional_header_dict.keys()
      pick_content_type = re.compile('header\s*content-type=(.*)')
      pick_url = re.compile('header\s*url=(.*)')
      for section in section_list:
        content_type_regex_str_match = pick_content_type.match(section)
        url_regex_str_match = pick_url.match(section)
        if content_type_regex_str_match is not None:
          content_type_regex_str = content_type_regex_str_match.group(1)
          content_type_regex = \
            re.compile('Content-Type:\s%s' % content_type_regex_str,
                       re.MULTILINE | re.IGNORECASE)
          if content_type_regex.search(fetched_data) is not None:
            for header, value in self.conditional_header_dict[section]:
              conditional_header_dict[header] = _formatConfiguration(value)
          continue
        if url_regex_str_match is not None:
          url_regex_str = url_regex_str_match.group(1)
          if re.compile(url_regex_str).match(url) is not None:
            for header, value in self.conditional_header_dict[section]:
              conditional_header_dict[header] = _formatConfiguration(value)
      return conditional_header_dict

    validator_dict = {}
    no_header_list = getNoCheckHeaderList(url, fetched_data)
    conditional_header_dict = getConditionalHeaderDict(url, fetched_data)
    conditional_header_list = conditional_header_dict.keys()
    global_header_list = self.header_list.keys()
    header_policy_list = []
    if conditional_header_list:
      conditional_check_header_set = (set(conditional_header_list)
                                      - set(no_header_list))
      for header in conditional_check_header_set:
        header_policy_list.append((header, conditional_header_dict[header]))
    else:
      global_check_header_set = (set(global_header_list)
                                 - set(no_header_list))
      for header in global_check_header_set:
        header_policy_list.append((header, self.header_list[header]))
    # return items
    return header_policy_list

  def _validateHeader(self, url, header, reference_value, fetched_data):
    """validate header with the header policy"""
    re_compiled = re.compile(self.generic_header_search_regex % header,
                             re.MULTILINE | re.IGNORECASE)
    match_object = re_compiled.search(fetched_data)
    if match_object is None:
      message = 'header:%r not found' % (header)
      self.report_dict.setdefault(url, []).append(message)
    else:
      read_value = match_object.group(1)
      if reference_value is True and not read_value:
        message = 'value of header:%r not found' % (header)
        self.report_dict.setdefault(url, []).append(message)
      elif isinstance(reference_value, (tuple,list)):
        if read_value not in reference_value:
          message = 'value of header:%r does not match'\
                    ' (%r not in %r)' %\
                                    (header, read_value, reference_value)
          self.report_dict.setdefault(url, []).append(message)

  def _isSameUrl(self, url):
    """
      Return whether the url is already checked or not.

      Example case):
      http://example.com/login_form
      http://example.com/login_form/
    """
    if url in (None, ''):
      return False
    same_url = None
    if url.endswith('/'):
      same_url = url.rstrip('/')
    else:
      same_url = '%s/' % url
    if same_url in self.report_dict:
      return True
    return False

  def _parseWgetLogs(self, wget_log_file, discarded_url_list=_MARKER,
                     prohibited_file_name_list=None,
                     prohibited_folder_name_list=None):
    """read wget logs and test Caching configuration
    """
    if discarded_url_list is _MARKER:
      first_pass = True
    else:
      first_pass = False
    x_varnish_reference_list = []

    for line in  wget_log_file.splitlines():
      logging.debug('wget output:%r' % line)
      code, value = self._parseWgetLine(line)
      if code == self.URL_CODE:
        # This is the first Line by URL checked in wget stdout
        url = value
        logging.debug('url:%r' % url)
        discarded = False
        if not first_pass and url in discarded_url_list:
          # URL already checked during first pass
          logging.debug('%r Discarded' % url)
          discarded = True
        elif self._isSameUrl(url):
          discarded = True
      if discarded:
        # keep reading wget process without doing anything
        continue
      if code == self.STATUS_CODE:
        if value not in self.ACCEPTABLE_STATUS_LIST:
          message = 'Page in error status:%r' % (value)
          self.report_dict.setdefault(url, []).append(message)
      if code == self.FILE_PATH_CODE:
        # Here we check if Response was cached
        file_path = os.path.sep.join((self.working_directory, value))
        folder_path , filename = os.path.split(file_path)
        if prohibited_file_name_list:
          if '?' in filename:
            filename = filename.rpartition('?')[0]
          if filename in prohibited_file_name_list:
            # Error
            message = '%r is prohibited as filename' % filename
            self.report_dict.setdefault(url, []).append(message)
        if prohibited_folder_name_list:
          for prohibited_folder_name in prohibited_folder_name_list:
            if prohibited_folder_name in folder_path.split(os.path.sep):
              message = '%r is prohibited as folder name' % prohibited_folder_name
              self.report_dict.setdefault(url, []).append(message)
              break
        file_object = None
        try:
          file_object = open(file_path, 'r')
        except IOError:
          # This is probably a folder try with /index.html
          index_file_path = file_path + os.path.sep + 'index.html'
          try:
            file_object = open(index_file_path, 'r')
          except IOError:
           # sometimes this is appended with .1
           number_file_path = file_path + '.1'
           try:
             file_object = open(number_file_path, 'r')
           except IOError:
             logging.info('File not found for url:%r %r' %\
                        (url, (file_path, index_file_path, number_file_path),))
             continue
        fetched_data = file_object.read()
        file_object.close()
        x_cache_header_match_object =\
                          self.x_cache_header_search_regex.search(fetched_data)
        if x_cache_header_match_object is None:
          # This RESPONSE is not cached by Varnish
          message = 'X-Cache header not found'
          self.report_dict.setdefault(url, []).append(message)
        else:
          # means X-Cache header is present in reponse
          # Read the X-Varnish header to know if backend has been touched
          x_varnish_match_object =\
                        self.x_varnish_header_search_regex.search(fetched_data)
          x_varnish_reference = x_varnish_match_object.group(1)
          logging.debug('x_varnish_reference:%r for url:%r' %\
                                                    (x_varnish_reference, url))
          hits = x_cache_header_match_object.group(1)
          if hits.isdigit():
            # This is a cached content with a positive hit value
            # Check if request didn't goes to the backend
            x_varnish_reference_list.append((x_varnish_reference, False, url))
            # dot not check this url in second pass
            logging.debug('will be discarded:%r' % url)
            discarded_url_list.append(url)
          else:
            x_varnish_reference_list.append((x_varnish_reference, True, url))
        # parse the web checker configuration file and run the header
        # validation method
        for header, reference_value in self._getHeaderPolicyList(
                                              url, fetched_data):
          self._validateHeader(url, header, reference_value, fetched_data)
    return x_varnish_reference_list, discarded_url_list[:]

  def start(self, prohibited_file_name_list=None,
            prohibited_folder_name_list=None):
    """Run test suite
    return errors
    """
    logging.info('#'*52)
    logging.info('date:%r' % (datetime.now().isoformat()))
    logging.info('#'*52)
    self._initFolder()
    logging.info('First pass:%r' % self.root_url)
    varnishlog_reading_process = self._runVarnishLog()
    wget_log_file = self._runSpider()
    varnishlog_data = self._readVarnishLog(varnishlog_reading_process)
    x_varnish_reference_list, discarded_url_list =\
                                             self._parseWgetLogs(wget_log_file)
    self._readVarnishLogAndGetIsBackendTouched(varnishlog_data,
                                                      x_varnish_reference_list)
    logging.info('End of First pass\n')
    [logging.debug(discarded_url) for discarded_url in discarded_url_list]
    self._initFolder()
    logging.info('Second pass:%r' % self.root_url)
    varnishlog_reading_process = self._runVarnishLog()
    wget_log_file = self._runSpider()
    varnishlog_data = self._readVarnishLog(varnishlog_reading_process)
    x_varnish_reference_list, discarded_url_list =\
                           self._parseWgetLogs(wget_log_file,
                       discarded_url_list=discarded_url_list,
                       prohibited_file_name_list=prohibited_file_name_list,
                       prohibited_folder_name_list=prohibited_folder_name_list)
    self._readVarnishLogAndGetIsBackendTouched(varnishlog_data,
                                                      x_varnish_reference_list)
    logging.info('End of second pass\n')
    if self.report_dict:
      report_message_list = ['*Errors*:']
      for url, message_list in six.iteritems(self.report_dict):
        unique_message_list = []
        for message in message_list:
          if message not in unique_message_list:
            unique_message_list.append(message)
        report_message_list.append('\n')
        report_message_list.append(url)
        report_message_list.extend(['\t%s' % message for message\
                                                       in unique_message_list])
      report_message = '\n'.join(report_message_list)
      signal = 'PROBLEM'
    else:
      report_message = 'No errors'
      signal = 'OK'
    subject = '%r:HTTP Cache checker results for %s' % (signal, self.root_url)
    if self.email_address:
      import smtplib
      message = 'Subject: %s\nFrom: %s\nTo: %s\n\n%s' %\
              (subject, self.email_address, self.email_address, report_message)
      server = smtplib.SMTP(self.smtp_host)
      server.sendmail(self.email_address, self.email_address, message)
      server.quit()
      return 'Email sended to %s' % self.email_address
    else:
      return subject + '\n' + report_message


from optparse import OptionParser
from six.moves import configparser

def _formatConfiguration(configuration):
  """ format the configuration"""
  if configuration in ('True', 'true', 'yes'):
    return True
  return configuration.splitlines()

def web_checker_utility():
  usage = "usage: %prog [options] config_path"
  parser = OptionParser(usage=usage)
  parser.add_option('-o', '--output_file',
                    dest='output_file')

  (options, args) = parser.parse_args()
  if len(args) != 1 :
    print(parser.print_help())
    parser.error('incorrect number of arguments')
  config_path = args[0]

  config = configparser.RawConfigParser()
  config.read(config_path)
  working_directory = config.get('web_checker', 'working_directory')
  url = config.get('web_checker', 'url')
  varnishlog_binary_path = config.get('web_checker', 'varnishlog_binary_path')
  wget_binary_path = 'wget'
  if config.has_option('web_checker', 'wget_binary_path'):
    wget_binary_path = config.get('web_checker', 'wget_binary_path')
  email_address = config.get('web_checker', 'email_address')
  smtp_host = config.get('web_checker', 'smtp_host')
  debug_level = config.get('web_checker', 'debug_level')
  file_log_path = 'web_checker.log'
  if config.has_option('web_checker', 'file_log_path'):
    file_log_path = config.get('web_checker', 'file_log_path')
  header_list = {}
  for header, configuration in config.items('header_list'):
    if header in config.defaults().keys():
      # defaults are shared for all sections.
      # so discard them from header_list
      continue
    value = _formatConfiguration(configuration)
    header_list[header] = value
  conditional_header_dict = {}
  no_header_dict = {}
  for section in config.sections():
    item_list = config.items(section)
    if re.compile("^header\s.*").match(section) is not None:
      conditional_header_dict.setdefault(section, []).extend(item_list)
    if re.compile("^no_header\s.*").match(section) is not None:
      no_header_dict.setdefault(section, []).extend(item_list)
  if config.has_section('erp5_extension_list'):
    prohibited_file_name_list = config.get('erp5_extension_list',
                                      'prohibited_file_name_list').splitlines()
    prohibited_folder_name_list = config.get('erp5_extension_list',
                                    'prohibited_folder_name_list').splitlines()
  else:
    prohibited_file_name_list = prohibited_folder_name_list = []
  instance = HTTPCacheCheckerTestSuite(url,
                                       working_directory,
                                       varnishlog_binary_path,
                                       wget_binary_path,
                                       header_list,
                                       email_address,
                                       smtp_host,
                                       debug_level,
                                       file_log_path,
                                       conditional_header_dict,
                                       no_header_dict)

  result = instance.start(prohibited_file_name_list=prohibited_file_name_list,
                       prohibited_folder_name_list=prohibited_folder_name_list)
  if options.output_file:
    with open(options.output_file, 'w') as file_object:
      file_object.write(result)
  else:
    print(result)


